# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from collections.abc import Sequence
import dataclasses
import enum
import functools
import json
import logging
import re
from typing import Any, AsyncIterable, Coroutine, Hashable, TypeVar

import bs4
from vertexai import generative_models
from vertexai import language_models
import numpy as np
import pandas as pd
import pydantic
import requests
from sklearn import cluster
from sklearn import neighbors
import tqdm

from copycat import google_ads


LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


class TqdmLogger:
  """File-like class redirecting tqdm progress bar to LOGGER."""

  def write(self, msg: str) -> None:
    LOGGER.info(msg.lstrip("\r"))

  def flush(self) -> None:
    pass


AsyncGenerationResponse = Coroutine[
    Any,
    Any,
    generative_models.GenerationResponse
    | AsyncIterable[generative_models.GenerationResponse],
]

GoogleAd = google_ads.GoogleAd
GoogleAdFormat = google_ads.GoogleAdFormat

ValidationError = pydantic.ValidationError
FinishReason = generative_models.FinishReason

SafetySettingsType = (
    dict[generative_models.HarmCategory, generative_models.HarmBlockThreshold]
    | list[generative_models.SafetySetting]
)

VECTORSTORE_PARAMS_FILE_NAME = "vectorstore_params.json"
VECTORSTORE_AD_EXEMPLARS_FILE_NAME = "vectorstore_ad_exemplars.csv"


class ModelName(enum.Enum):
  GEMINI_1_0_PRO = "gemini-pro"
  GEMINI_1_5_PRO_PREVIEW = "gemini-1.5-pro-preview-0514"
  GEMINI_1_5_PRO_001 = "gemini-1.5-pro-001"
  GEMINI_1_5_PRO_002 = "gemini-1.5-pro-002"
  GEMINI_1_5_PRO = "gemini-1.5-pro-002"
  GEMINI_1_5_FLASH_PREVIEW = "gemini-1.5-flash-preview-0514"
  GEMINI_1_5_FLASH_001 = "gemini-1.5-flash-001"
  GEMINI_1_5_FLASH_002 = "gemini-1.5-flash-002"
  GEMINI_1_5_FLASH = "gemini-1.5-flash-002"


class EmbeddingModelName(enum.Enum):
  TEXT_EMBEDDING = "text-embedding-004"
  TEXT_MULTILINGUAL_EMBEDDING = "text-multilingual-embedding-002"


class ExemplarSelectionMethod(enum.Enum):
  AFFINITY_PROPAGATION = "affinity_propagation"
  RANDOM = "random"


class TextGenerationRequest(pydantic.BaseModel):
  """The request to generate text."""

  keywords: str
  existing_ad_copy: GoogleAd
  system_instruction: str
  prompt: list[generative_models.Content]
  chat_model_name: ModelName
  temperature: float
  top_k: int
  top_p: float
  safety_settings: SafetySettingsType | None

  class Config:
    arbitrary_types_allowed = True

  def to_markdown(self):
    lines = [
        "**Keywords:**",
        self.keywords,
    ]

    if self.existing_ad_copy.headlines:
      lines.extend(
          ["**Existing headlines:**", f"{self.existing_ad_copy.headlines}"]
      )
    if self.existing_ad_copy.descriptions:
      lines.extend([
          "**Existing descriptions:**",
          f"{self.existing_ad_copy.descriptions}",
      ])

    lines.extend([
        "**Model Parameters:**",
        f"Model name: {self.chat_model_name.value}",
        f"Temperature: {self.temperature}",
        f"Top K: {self.top_k}",
        f"Top P: {self.top_p}",
        f"Safety settings: {self.safety_settings}",
        "**System instruction:**",
        self.system_instruction,
    ])

    for content in self.prompt:
      lines.append(f"**{content.role.title()}:**")
      lines.append(content.parts[0].text)

    return "\n\n".join(lines)


class ExampleAd(pydantic.BaseModel):
  """An example ad.

  Attributes:
    google_ad: The google ad containing the headlines and descriptions.
    keywords: The keywords this ad was used for.
  """

  google_ad: GoogleAd
  keywords: str

  @classmethod
  def from_flat_values(
      cls, keywords: str, headlines: list[str], descriptions: list[str]
  ) -> "ExampleAd":
    """Creates an ExampleAd from keywords, headlines and descriptions."""
    return cls(
        google_ad=GoogleAd(headlines=headlines, descriptions=descriptions),
        keywords=keywords,
    )


@dataclasses.dataclass
class AdCopyVectorstore:
  """The vector store containing the ad copies.

  Each record contains both a text that will be matched to queries and some
  metadata. The text is either a headline or a description that exists in the
  ad, and the metadata contains the full list of headlines, descriptions and
  keywords for that ad. Each ad will appear in the vectorstore multiple times,
  once for each headline and description it uses. This allows the ads to be
  matched to the query based on the most relavent individual headline or
  description, rather than an average over all of them.

  Attributes:
    embedding_model_name: The name of the embedding model to use.
    ad_exemplars: The example ads available to be used as in context examples.
    dimensionality: The dimensionality of the embedding model.
    embeddings_batch_size: The batch size to use when generating embeddings.
    unique_headlines: The unique headlines in the vectorstore.
    unique_descriptions: The unique descriptions in the vectorstore.
    n_exemplars: The total number of exemplars in the vectorstore.
  """

  embedding_model_name: EmbeddingModelName
  ad_exemplars: pd.DataFrame
  dimensionality: int
  embeddings_batch_size: int

  @classmethod
  def _generate_embeddings(
      cls,
      texts: list[str],
      *,
      embedding_model_name: EmbeddingModelName,
      dimensionality: int,
      batch_size: int,
      task_type: str,
      progress_bar: bool = False,
  ) -> list[list[float]]:
    """Generates embeddings for the provided texts.

    Args:
      texts: The texts to generate embeddings for.
      embedding_model_name: The name of the embedding model to use.
      dimensionality: The dimensionality of the embedding model.
      batch_size: The batch size to use when generating embeddings.
      task_type: The task type to use when generating embeddings.
      progress_bar: Whether to show a progress bar.

    Returns:
      The generated embeddings.
    """
    embedding_model = language_models.TextEmbeddingModel.from_pretrained(
        embedding_model_name.value
    )
    n_batches = np.ceil(len(texts) / batch_size)
    LOGGER.debug(
        "Using embedding model: %s, dimensionality: %d, task type: %s",
        embedding_model_name.value,
        dimensionality,
        task_type,
    )
    LOGGER.debug(
        "Generating %d embeddings in %d batches.", len(texts), n_batches
    )

    embeddings = []

    texts_batch_iterator = np.array_split(texts, n_batches)
    if progress_bar:
      texts_batch_iterator = tqdm.tqdm(
          texts_batch_iterator,
          desc="Generating embeddings",
          file=TqdmLogger(),
          mininterval=5,
      )

    for texts_batch in texts_batch_iterator:
      embedding_inputs = [
          language_models.TextEmbeddingInput(ad_markdown, task_type)
          for ad_markdown in texts_batch
      ]
      embedding_outputs = embedding_model.get_embeddings(
          embedding_inputs, output_dimensionality=dimensionality
      )
      embeddings.extend([emb.values for emb in embedding_outputs])

    LOGGER.debug("Embeddings generated.")
    return embeddings

  def embed_documents(self, texts: list[str]) -> list[list[float]]:
    """Generates embeddings for the provided texts."""
    return self._generate_embeddings(
        texts,
        embedding_model_name=self.embedding_model_name,
        dimensionality=self.dimensionality,
        batch_size=self.embeddings_batch_size,
        task_type="RETRIEVAL_DOCUMENT",
        progress_bar=False,
    )

  def embed_queries(self, texts: list[str]) -> list[list[float]]:
    """Generates embeddings for the provided texts."""
    return self._generate_embeddings(
        texts,
        embedding_model_name=self.embedding_model_name,
        dimensionality=self.dimensionality,
        batch_size=self.embeddings_batch_size,
        task_type="RETRIEVAL_QUERY",
        progress_bar=False,
    )

  @classmethod
  def _get_exemplars(
      cls,
      data: pd.DataFrame,
      *,
      embeddings_column: str,
      affinity_preference: float | None,
      max_exemplars: int,
  ) -> pd.DataFrame:
    """Uses Affinity Propagation to find exemplar ads."""
    LOGGER.info("Getting exemplars with Affinity Propagation.")
    embeddings = np.asarray(data[embeddings_column].values.tolist())

    clusterer = cluster.AffinityPropagation(preference=affinity_preference)
    clusterer.fit(embeddings)
    exemplars = (
        data.iloc[clusterer.cluster_centers_indices_]
        .copy()
        .reset_index(drop=True)
    )

    if len(exemplars) > max_exemplars:
      LOGGER.info(
          "Affinity Propagation returned too many exemplars (%d). Sampling to"
          " get %d.",
          len(exemplars),
          max_exemplars,
      )
      exemplars = exemplars.sample(max_exemplars)

    LOGGER.info("Got %d exemplars with Affinity Propagation.", len(exemplars))
    return exemplars

  @classmethod
  def _deduplicate_ads(cls, data: pd.DataFrame) -> pd.DataFrame:
    """Deduplicates the ads in the training data.

    If the same ads are used for multiple sets of keywords, select just one
    random keywords set for each ad. We don't need to have identical ads in
    the vectorstore.

    Args:
      data: The training data containing the headlines, descriptions and
        keywords.

    Returns:
      The deduplicated training data.
    """
    n_starting_rows = len(data)
    LOGGER.info("Deduplicating ads (starting with %d rows).", n_starting_rows)

    data = data.copy()
    data["headlines"] = data["headlines"].apply(tuple)
    data["descriptions"] = data["descriptions"].apply(tuple)
    data = (
        data.groupby(["headlines", "descriptions"], group_keys=False)
        .sample(1)
        .reset_index(drop=True)
    )
    data["headlines"] = data["headlines"].apply(list)
    data["descriptions"] = data["descriptions"].apply(list)

    n_deduped_rows = len(data)
    LOGGER.info(
        "Deduplication removed %d rows.", n_starting_rows - n_deduped_rows
    )

    return data

  @classmethod
  def create_from_pandas(
      cls,
      training_data: pd.DataFrame,
      *,
      embedding_model_name: str | EmbeddingModelName,
      dimensionality: int,
      max_initial_ads: int,
      max_exemplar_ads: int,
      affinity_preference: float | None,
      embeddings_batch_size: int,
      exemplar_selection_method: (
          str | ExemplarSelectionMethod
      ) = "affinity_propagation",
  ) -> "AdCopyVectorstore":
    """Creates a vector store containing the ad copies from pandas.

    The vectorstore is created from the provided training data. The training
    data contains the real ad copies and keywords they were used for. Make sure
    the ad copy is high quality as this is what the model will learn from.

    The training_data must contain the following columns:
      - headlines: The headlines of the ad copy. This should be a list of
        strings.
      - descriptions: The descriptions of the ad copy. This should be a list of
        strings.
      - keywords: The keywords the ad copy was used for. This should be a
        string of comma separated keywords.

    The vectorstore is created by:
      1.  Deduplicating the ads in the training data. This ensures that each ad
          is only used once in the vectorstore.
      2.  Sampling the training data to a maximum of max_initial_ads. This
          ensures that the next steps are not too slow.
      3.  Generating embeddings for the ads. This is done using the provided
          embedding model name.
      4.  Applying affinity propogation to find "exemplar ads", which are
          ads that are representative of the training data, but are not too
          similar to each other.
      5.  Sampling the exemplar ads to a maximum of max_exemplar_ads. This
          ensures that the vectorstore does not become too large.

    The affinity propogation algorithm depends on the affinity_preference
    parameter. A higher affinity_preference will result in more exemplar ads
    being selected, while a lower affinity_preference will result in fewer
    exemplar ads being selected. The affinity preference should be a negative
    number. If set to None it automatically selects the number of
    exemplar ads based on the data.

    Args:
      training_data: The training data containing the real ad copies and
        keywords.
      embedding_model_name: The name of the embedding model to use.
      dimensionality: The dimensionality of the embedding model.
      max_initial_ads: The maximum number of ads to use from the training data.
        This is used to speed up the process of creating the vectorstore.
      max_exemplar_ads: The maximum number of exemplar ads to use in the
        vectorstore.
      affinity_preference: The affinity preference to use when finding exemplar
        ads.
      embeddings_batch_size: The batch size to use when generating embeddings.
      exemplar_selection_method: The method to use to select the exemplar ads.
        Either "affinity_propagation" or "random". Defaults to
        "affinity_propagation".

    Returns:
      An instance of the AdCopyVectorstore containing the exemplar ads.

    Raises:
      ValueError: If the exemplar selection method is not supported.
    """
    embedding_model_name = EmbeddingModelName(embedding_model_name)
    exemplar_selection_method = ExemplarSelectionMethod(
        exemplar_selection_method
    )
    LOGGER.info(
        "Creating AdCopyVectorstore from %d rows of training data.",
        len(training_data),
    )

    data = (
        training_data[["headlines", "descriptions", "keywords"]]
        .copy()
        .pipe(cls._deduplicate_ads)
    )

    if len(data) > max_initial_ads:
      LOGGER.info("Sampling %d ads from %d ads.", max_initial_ads, len(data))
      data = data.sample(max_initial_ads)

    data["ad_markdown"] = data.apply(lambda x: str(GoogleAd(**x)), axis=1)
    LOGGER.info(
        "Finding exemplar ads with method = %s.",
        exemplar_selection_method.value,
    )
    if (
        exemplar_selection_method
        is ExemplarSelectionMethod.AFFINITY_PROPAGATION
    ):
      data["embeddings"] = cls._generate_embeddings(
          data["ad_markdown"].values.tolist(),
          embedding_model_name=embedding_model_name,
          dimensionality=dimensionality,
          batch_size=embeddings_batch_size,
          task_type="RETRIEVAL_DOCUMENT",
          progress_bar=True,
      )

      ad_exemplars = cls._get_exemplars(
          data,
          embeddings_column="embeddings",
          affinity_preference=affinity_preference,
          max_exemplars=max_exemplar_ads,
      )
    elif exemplar_selection_method is ExemplarSelectionMethod.RANDOM:
      if len(data) > max_exemplar_ads:
        ad_exemplars = data.sample(max_exemplar_ads)
      else:
        ad_exemplars = data

      ad_exemplars["embeddings"] = cls._generate_embeddings(
          ad_exemplars["ad_markdown"].values.tolist(),
          embedding_model_name=embedding_model_name,
          dimensionality=dimensionality,
          batch_size=embeddings_batch_size,
          task_type="RETRIEVAL_DOCUMENT",
          progress_bar=True,
      )

    else:
      LOGGER.error(
          "Unsupported exemplar selection method: %s",
          exemplar_selection_method.value,
      )
      raise ValueError(
          f"Unsupported exemplar selection method: {exemplar_selection_method}"
      )

    LOGGER.info(
        "Reduced %d total ads to %d exemplar ads.",
        len(training_data),
        len(ad_exemplars),
    )

    return cls(
        embedding_model_name=embedding_model_name,
        ad_exemplars=ad_exemplars,
        dimensionality=dimensionality,
        embeddings_batch_size=embeddings_batch_size,
    )

  @classmethod
  def from_dict(cls, params: dict[str, Any]) -> "AdCopyVectorstore":
    """Loads the vectorstore from the provided dict.

    Schema of params:
      embedding_model_name: The name of the embedding model to use.
      ad_exemplars: The ad exemplars to use in the vectorstore, as a pandas
        dataframe that has been converted to a dict with to_dict("tight").
      dimensionality: The dimensionality of the embedding model.
      embeddings_batch_size: The batch size to use when generating embeddings.

    Args:
      params: The dict containing the parameters to use to create the
        AdCopyVectorstore. See above for the schema.

    Returns:
      An instance of the AdCopyVectorstore.

    Raises:
      KeyError: If any of the required keys are missing from the dict.
    """
    required_keys = {
        "embedding_model_name",
        "ad_exemplars",
        "dimensionality",
        "embeddings_batch_size",
    }
    missing_keys = required_keys - set(params.keys())
    if missing_keys:
      LOGGER.error("Missing required keys: %s", missing_keys)
      raise KeyError(f"Missing required keys: {missing_keys}")

    params = params.copy()
    params["embedding_model_name"] = EmbeddingModelName(
        params["embedding_model_name"]
    )
    params["ad_exemplars"] = pd.DataFrame.from_dict(
        params["ad_exemplars"], orient="tight"
    )
    return cls(**params)

  @classmethod
  def from_json(cls, json_string: str) -> "AdCopyVectorstore":
    """Loads the vectorstore from the provided json string.

    Schema of params:
      embedding_model_name: The name of the embedding model to use.
      ad_exemplars: The ad exemplars to use in the vectorstore, as a pandas
        dataframe that has been converted to a dict with to_dict("tight").
      dimensionality: The dimensionality of the embedding model.
      embeddings_batch_size: The batch size to use when generating embeddings.

    Args:
      json_string: The json string containing the parameters to use to create
        the AdCopyVectorstore. See above for the schema.

    Returns:
      An instance of the AdCopyVectorstore.

    Raises:
      KeyError: If any of the required keys are missing from the dict.
    """
    return cls.from_dict(json.loads(json_string))

  def to_dict(self) -> dict[str, Any]:
    """Serializes the vectorstore to a dict."""
    return {
        "embedding_model_name": self.embedding_model_name.value,
        "dimensionality": self.dimensionality,
        "embeddings_batch_size": self.embeddings_batch_size,
        "ad_exemplars": self.ad_exemplars.to_dict(orient="tight"),
    }

  def to_json(self) -> str:
    """Serializes the vectorstore to a json string."""
    return json.dumps(self.to_dict())

  @functools.cached_property
  def nearest_neighbors(self) -> neighbors.NearestNeighbors:
    """The nearest neighbors model used to find similar ads."""
    embeddings = np.asarray(self.ad_exemplars["embeddings"].values.tolist())
    model = neighbors.NearestNeighbors()
    model.fit(embeddings)
    return model

  @functools.cached_property
  def unique_headlines(self) -> set[str]:
    return set(self.ad_exemplars["headlines"].explode().unique().tolist())

  @functools.cached_property
  def unique_descriptions(self) -> set[str]:
    return set(self.ad_exemplars["descriptions"].explode().unique().tolist())

  @property
  def n_exemplars(self) -> int:
    """The total number of exemplars in the vectorstore."""
    return len(self.ad_exemplars)

  def get_relevant_ads_and_embeddings_from_embeddings(
      self,
      query_embeddings: list[list[float]],
      k: int,
  ) -> tuple[list[list[ExampleAd]], list[list[float]]]:
    """Gets the most relevant ads and their embeddings for the query embeddings.

    Args:
      query_embeddings: The list of query embeddings to use to retrieve the ads.
        These are typically the embeddings of the keywords used to generate the
        ad copy.
      k: The number of ads to return for each query.

    Returns:
      The k most relevant ads and their embeddings for each query as two lists.
    """
    k = min(self.n_exemplars, k)

    similar_ad_ids = self.nearest_neighbors.kneighbors(
        query_embeddings, n_neighbors=k, return_distance=False
    )
    similar_ads = [
        list(
            map(
                lambda x: ExampleAd.from_flat_values(**x),
                self.ad_exemplars.iloc[ids][
                    ["headlines", "descriptions", "keywords"]
                ].to_dict("records"),
            )
        )
        for ids in similar_ad_ids
    ]
    similar_ad_embeddings = [
        self.ad_exemplars.iloc[ids]["embeddings"].values.tolist()
        for ids in similar_ad_ids
    ]
    return similar_ads, similar_ad_embeddings

  def get_relevant_ads(
      self, queries: list[str], k: int
  ) -> list[list[ExampleAd]]:
    """Returns the k most relevant ads for the provided query.

    Args:
      queries: The list of queries to use to retrieve the ads. These are
        typically the keywords used to generate the ad copy.
      k: The number of ads to return for each query.

    Returns:
      The k most relavent ads for each query
    """
    query_embeddings = self.embed_queries(queries)
    relevant_ads, _ = self.get_relevant_ads_and_embeddings_from_embeddings(
        query_embeddings, k
    )
    return relevant_ads


def _construct_instruction_for_number_of_headlines_and_descriptions(
    existing_ad_copy: GoogleAd,
    ad_format: GoogleAdFormat,
) -> str:
  """Returns the instruction with how many headlines / descriptions to generate.

  If the ad already has some headlines and descriptions, then the prompt needs
  to inclide them and explain to Copycat that it should just generate the
  missing headlines and descriptions.

  If the existing headlines and descriptions are empty then it just needs to
  explain the number of headlines and descriptions required for the format.

  If the ad is already complete, then it should raise an error.

  Args:
    existing_ad_copy: The existing headlines and descriptions.
    ad_format: The ad format to generate.

  Returns:
    The instruction for extending the ad copy.

  Raises:
    ValueError: If the ad is already complete, meaning there is nothing to
      generate.
  """
  n_existing_headlines = len(existing_ad_copy.headlines)
  n_existing_descriptions = len(existing_ad_copy.descriptions)
  n_required_headlines = ad_format.max_headlines - n_existing_headlines
  n_required_descriptions = ad_format.max_descriptions - n_existing_descriptions

  no_headlines = n_existing_headlines == 0
  no_descriptions = n_existing_descriptions == 0
  complete_headlines = n_existing_headlines >= ad_format.max_headlines
  complete_descriptions = n_existing_descriptions >= ad_format.max_descriptions

  # If the ad is already complete, then there is nothing to generate.
  if complete_headlines and complete_descriptions:
    LOGGER.error("Trying to generate an ad that is already complete.")
    raise ValueError("Trying to generate an ad that is already complete.")

  # If the ad is empty, then we need to generate the required headlines and
  # descriptions.
  if no_headlines and no_descriptions:
    return (
        f"Please write {n_required_headlines} headlines and"
        f" {n_required_descriptions} descriptions for this ad."
    )

  instructions = []
  # First, if there are already some headlines and descriptions, explain this
  # to Copycat.
  if not no_headlines and not no_descriptions:
    instructions.extend([
        (
            f"This ad already has {n_existing_headlines} headlines and"
            f" {n_existing_descriptions} descriptions:\n"
        ),
        f"- headlines: {existing_ad_copy.headlines}",
        f"- descriptions: {existing_ad_copy.descriptions}\n",
    ])
  elif not no_headlines:
    instructions.extend([
        f"This ad already has {n_existing_headlines} headlines.\n",
        f"- headlines: {existing_ad_copy.headlines}\n",
    ])
  elif not no_descriptions:
    instructions.extend([
        f"This ad already has {n_existing_descriptions} descriptions.\n",
        f"- descriptions: {existing_ad_copy.descriptions}\n",
    ])

  # If the ad still needs to generate more headlines and descriptions, make it
  # clear how many are required.
  if not complete_headlines and not complete_descriptions:
    instructions.append(
        f"Please write {n_required_headlines} more headlines and"
        f" {n_required_descriptions} more descriptions to complete this ad."
    )
  elif not complete_headlines and complete_descriptions:
    instructions.append(
        f"Please write {n_required_headlines} more headlines to complete this"
        " ad. You do not need to write any descriptions, as there are enough"
        " already."
    )
  elif not complete_descriptions and complete_headlines:
    instructions.append(
        f"Please write {n_required_descriptions} more descriptions to complete"
        " this ad. You do not need to write any headlines, as there are enough"
        " already."
    )

  return "\n".join(instructions)


def _construct_new_ad_copy_user_message(
    keywords: str,
    ad_format: GoogleAdFormat,
    existing_ad_copy: GoogleAd | None = None,
    keywords_specific_instructions: str = "",
) -> generative_models.Content:
  """Returns the user message for generating new ad copy.

  Args:
    keywords: The keywords to generate the ad copy for.
    ad_format: The ad format to generate.
    existing_ad_copy: The existing headlines and descriptions for this ad. The
      prompt will ask Copycat to extend this ad copy to the required number of
      headlines and descriptions for the ad format. If None, then the ad copy is
      empty.
    keywords_specific_instructions: Any additional context to use for the new
      keywords. This could include things like information from the landing
      page, information about specific discounts or promotions, or any other
      relevant information.
  """
  if existing_ad_copy is None:
    existing_ad_copy = GoogleAd(headlines=[], descriptions=[])

  content = ""
  if keywords_specific_instructions:
    content += (
        "For the next set of keywords, please consider the following additional"
        f" instructions:\n\n{keywords_specific_instructions}\n\n"
    )
  content += _construct_instruction_for_number_of_headlines_and_descriptions(
      existing_ad_copy=existing_ad_copy,
      ad_format=ad_format,
  )
  content += f"\n\nKeywords: {keywords}"

  return generative_models.Content(
      role="user",
      parts=[generative_models.Part.from_text(content)],
  )


def construct_system_instruction(
    system_instruction: str,
    style_guide: str,
    system_instruction_kwargs: dict[str, Any],
) -> str:
  """Constructs the system instruction by adding the style guide and kwargs.

  Args:
    system_instruction: The system instruction to use. This should explain the
      task to the model.
    style_guide: The style guide to use.
    system_instruction_kwargs: The keyword arguments are used to replace any
      placeholders in the system prompt.

  Returns:
  The formatted system prompt.
  """
  if system_instruction_kwargs:
    system_instruction = system_instruction.format(**system_instruction_kwargs)
  if style_guide:
    system_instruction += "\n\n" + style_guide
  return system_instruction


def construct_new_ad_copy_prompt(
    example_ads: list[ExampleAd],
    keywords: str,
    ad_format: GoogleAdFormat,
    existing_ad_copy: GoogleAd | None = None,
    keywords_specific_instructions: str = "",
) -> list[generative_models.Content]:
  """Constructs the full copycat prompt for generating new ad copy.

  The prompt consists of a list of in-context examples for new ad copy
  generation. This is a list of messages, alternating between the keywords and
  expected response from each example ad. The expected response is a json string
  containing the headlines and descriptions of the ad copy. The messages are
  sorted so that the most relevant examples are last. This ensures the model
  see's the most relevant examples last, making them more likely to influence
  the model's output. The final message contains the keywords to generate the ad
  copy for, and the additional context for the new keywords from the
  keywords_specific_instructions if it exists.

  Args:
    example_ads: The list of example ads to use as in-context examples.
    keywords: The keywords to generate the ad copy for.
    ad_format: The ad format to generate.
    existing_ad_copy: The existing headlines and descriptions for this ad. The
      prompt will ask Copycat to extend this ad copy to the required number of
      headlines and descriptions for the ad format. If None, then the ad copy is
      empty.
    keywords_specific_instructions: Any additional context to use for the new
      keywords. This could include things like information from the landing
      page, information about specific discounts or promotions, or any other
      relevant information.

  Returns:
    A list of Content representing the prompt.
  """
  prompt = []
  for example in reversed(example_ads):
    example_ad_format = ad_format.model_copy(
        update={
            "max_headlines": example.google_ad.headline_count,
            "max_descriptions": example.google_ad.description_count,
        }
    )
    prompt.append(
        _construct_new_ad_copy_user_message(
            example.keywords, ad_format=example_ad_format
        )
    )
    prompt.append(
        generative_models.Content(
            role="model",
            parts=[
                generative_models.Part.from_text(
                    example.google_ad.model_dump_json()
                )
            ],
        )
    )

  prompt.append(
      _construct_new_ad_copy_user_message(
          keywords,
          ad_format=ad_format,
          existing_ad_copy=existing_ad_copy,
          keywords_specific_instructions=keywords_specific_instructions,
      )
  )

  return prompt


HashableTypeVar = TypeVar("HashableTypeVar", bound=Hashable)


def _deduplicate_list_keep_order(
    seq: Sequence[HashableTypeVar],
) -> list[HashableTypeVar]:
  seen = set()
  seen_add = seen.add
  return [x for x in seq if not (x in seen or seen_add(x))]


def remove_invalid_headlines_and_descriptions(
    google_ad: GoogleAd, google_ad_format: GoogleAdFormat
) -> None:
  """Removes invalid headlines and descriptions from the ad.

  First it removes any duplicate headlines or descriptions, then removes any
  headlines or descriptions that are too long. Then it removes any headlines or
  descriptions that are not in the first k headlines or descriptions.

  Args:
    google_ad: The ad to remove the invalid headlines and descriptions from.
    google_ad_format: The format of the ad.
  """
  google_ad.headlines = _deduplicate_list_keep_order(google_ad.headlines)
  google_ad.descriptions = _deduplicate_list_keep_order(google_ad.descriptions)

  google_ad.headlines = [
      headline
      for headline in google_ad.headlines
      if len(google_ads.parse_google_ads_special_variables(headline))
      <= google_ad_format.max_headline_length
  ]
  google_ad.descriptions = [
      description
      for description in google_ad.descriptions
      if len(google_ads.parse_google_ads_special_variables(description))
      <= google_ad_format.max_description_length
  ]

  if len(google_ad.headlines) > google_ad_format.max_headlines:
    google_ad.headlines = google_ad.headlines[: google_ad_format.max_headlines]
  if len(google_ad.descriptions) > google_ad_format.max_descriptions:
    google_ad.descriptions = google_ad.descriptions[
        : google_ad_format.max_descriptions
    ]


def async_generate_google_ad_json(
    request: TextGenerationRequest,
) -> AsyncGenerationResponse:
  """Generates a GoogleAd from the text generation request asynchronously.

  This function ensures that the generated response is a valid json
  representation of a GoogleAd, by appending formatting instructions to the
  system instruction and including a response schema in the generation config
  for models that accept it.

  Args:
    request: The text generation request, containing the prompt, system
      instruction, style guide, and other parameters.

  Returns:
    The generated response, which is a valid json representation of a GoogleAd.
  """
  generation_config_params = dict(
      temperature=request.temperature,
      top_k=request.top_k,
      top_p=request.top_p,
      response_mime_type="application/json",
  )

  generation_config_params["response_schema"] = {
      "type": "OBJECT",
      "properties": {
          "headlines": {
              "type": "ARRAY",
              "items": {
                  "type": "string",
                  "description": (
                      "The headlines for the ad. Must be fewer than 30"
                      " characters."
                  ),
              },
          },
          "descriptions": {
              "type": "ARRAY",
              "items": {
                  "type": "string",
                  "description": (
                      "The descriptions for the ad. Must be fewer than 90"
                      " characters."
                  ),
              },
          },
      },
      "required": ["headlines", "descriptions"],
  }

  generation_config = generative_models.GenerationConfig(
      **generation_config_params
  )

  LOGGER.debug("System instruction: %s", request.system_instruction)
  LOGGER.debug("Prompt: %s", request.prompt)

  model = generative_models.GenerativeModel(
      model_name=request.chat_model_name.value,
      generation_config=generation_config,
      system_instruction=request.system_instruction,
      safety_settings=request.safety_settings,
  )

  response = model.generate_content_async(request.prompt)

  return response


def generate_google_ad_json_batch(
    requests: list[TextGenerationRequest],
) -> list[generative_models.GenerationResponse]:
  """Generates a GoogleAd from the provided text generation request.

  This function ensures that the generated response is a valid json
  representation of a GoogleAd, by appending formatting instructions to the
  system instruction and including a response schema in the generation config
  for models that accept it.

  Args:
    requests: A list of text generation requests, containing the prompts, system
      instructions, style guides, and other parameters.

  Returns:
    The generated responses, which are valid json representations of GoogleAds.

  Raises:
    RuntimeError: If one of the responses is not a valid json representation of
    a GoogleAd. This shouldn't happen unless the gemini api changes.
  """
  try:
    loop = asyncio.get_running_loop()
  except RuntimeError:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

  outputs = loop.run_until_complete(
      asyncio.gather(*list(map(async_generate_google_ad_json, requests)))
  )
  for output in outputs:
    if not isinstance(output, generative_models.GenerationResponse):
      LOGGER.error(
          "One of the responses is not a GenerationResponse. Instead got: %s",
          output,
      )
      raise RuntimeError(
          "One of the responses is not a GenerationResponse. Instead got:"
          f" {output}"
      )

  return outputs


def extract_urls_for_keyword_instructions(
    keyword_instructions: list[str],
) -> list[str]:
  """Extracts pages from keyword instructions.

  Args:
      keyword_instructions: keyword instructions or a string that may contain a
        URL.

  Returns:
      list of keyword instructions with retrieved web page content if a URL is
      found and successfully fetched.
  """
  pages = []
  for instruction in keyword_instructions:
    url = extract_url_from_string(instruction)
    if url and is_valid_url(url):
      try:
        response = requests.get(url)
        response.raise_for_status()
        soup = bs4.BeautifulSoup(response.content, "html.parser")
        if len(instruction) == len(url):  # if the instruction is only the url
          pages.append(f"Web page content of {url}: {soup.get_text()}")
        else:
          pages.append(f"{instruction} ## Content of {url}: {soup.get_text()}")
      except requests.exceptions.RequestException as e:
        pages.append(
            instruction
        )  # Keep the original instruction if there's an error
    else:
      pages.append(instruction)
  return pages


def extract_url_from_string(text: str) -> str | None:
  """Extracts a URL from a string.

  Args:
      text: The string that may contain a URL.

  Returns:
      The extracted URL if found, otherwise None.
  """
  url_pattern = re.compile(
      r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
  )
  match = url_pattern.search(text)
  return match.group() if match else None


def is_valid_url(url: str) -> bool:
  """Checks if a url is valid.

  Args:
      url: to check if valid url.

  Returns:
      Boolean dependent on if it is a valid url
  """
  regex = re.compile(
      r"^(https?):\/\/" r"([a-zA-Z0-9.-]+)\." r"([a-zA-Z]{2,})" r"(\/[^\s]*)?$",
      re.IGNORECASE,
  )

  return re.fullmatch(regex, url) is not None



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""States for the Copycat UI."""
import dataclasses
import logging

import mesop as me

from copycat import ad_copy_generator
from copycat import google_ads


DEFAULT_GCP_PROJECT_ID = ""


def set_default_gcp_project_id(project_id: str) -> None:
  """Sets the default GCP project ID."""
  global DEFAULT_GCP_PROJECT_ID
  DEFAULT_GCP_PROJECT_ID = project_id


@me.stateclass
class AppState:
  highlighted_url: str = "/setup"
  new_google_sheet_name: str = ""
  new_google_sheet_url: str = ""
  new_google_sheet_include_demo_data: bool = True
  google_sheet_name: str = ""
  google_sheet_url: str = ""
  display_google_sheet: bool = False
  show_starting_dialog: bool = True
  google_sheet_is_valid: bool = False
  has_copycat_instance: bool = False
  new_ad_preview_request: str = ""
  log_level: int = logging.INFO
  show_copycat_instance_created_snackbar: bool = False
  show_ad_copy_generated_snackbar: bool = False


@me.stateclass
class CopycatParamsState:
  vertex_ai_project_id: str = dataclasses.field(
      default_factory=lambda: DEFAULT_GCP_PROJECT_ID
  )
  vertex_ai_location: str = "us-central1"

  company_name: str = ""
  ad_format: str = "responsive_search_ad"
  max_headlines: int = google_ads.get_google_ad_format(
      "responsive_search_ad"
  ).max_headlines
  max_descriptions: int = google_ads.get_google_ad_format(
      "responsive_search_ad"
  ).max_descriptions
  language: str = "English"
  embedding_model_name: str = (
      ad_copy_generator.EmbeddingModelName.TEXT_EMBEDDING.value
  )
  how_to_handle_special_variables: str = "replace"
  on_invalid_ad: str = "drop"
  embedding_model_dimensionality: float = 768
  embedding_model_batch_size: int = 50
  max_initial_ads: int = 2000
  max_exemplar_ads: int = 200
  exemplar_selection_method: str = (
      ad_copy_generator.ExemplarSelectionMethod.AFFINITY_PROPAGATION.value
  )
  custom_affinity_preference: float = -0.5
  use_custom_affinity_preference: bool = False
  min_ad_strength: str = "POOR"

  style_guide_chat_model_name: str = (
      ad_copy_generator.ModelName.GEMINI_1_5_PRO.value
  )
  style_guide_temperature: float = 0.95
  style_guide_top_k: int = 40
  style_guide_top_p: float = 0.95
  style_guide_files_uri: str = ""
  style_guide_additional_instructions: str = ""
  style_guide_use_exemplar_ads: bool = True
  style_guide: str = ""
  generated_style_guide: str = ""

  new_ads_use_style_guide: bool = True
  new_ads_num_in_context_examples: int = 5
  new_ads_chat_model_name: str = (
      ad_copy_generator.ModelName.GEMINI_1_5_FLASH.value
  )
  new_ads_temperature: float = 0.95
  new_ads_top_k: int = 40
  new_ads_top_p: float = 0.95
  new_ads_allow_memorised_headlines: bool = False
  new_ads_allow_memorised_descriptions: bool = False
  new_ads_batch_size: int = 15
  new_ads_generation_limit: int = 30
  new_ads_fill_gaps: bool = True
  new_ads_number_of_versions: int = 1



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Models used for testing, which don't require a real VertexAI call."""

from typing import Any, Callable

from vertexai import language_models
from vertexai.generative_models import _generative_models as generative_models
import mock
import numpy as np
import pandas as pd

from copycat import ad_copy_generator
from copycat import copycat


GenerationConfigType = generative_models.GenerationConfigType
SafetySettingsType = generative_models.SafetySettingsType
PartsType = generative_models.PartsType
Tool = generative_models.Tool
ToolConfig = generative_models.ToolConfig
GenerationResponse = generative_models.GenerationResponse
ContentsType = generative_models.ContentsType


def _random_embedding(
    text: language_models.TextEmbeddingInput, output_dimensionality: int
) -> language_models.TextEmbedding:
  seed = int.from_bytes(
      (text.text + text.task_type).encode("utf-8"), byteorder="big"
  )
  embeddings = (
      np.random.Generator(np.random.MT19937(seed))
      .normal(size=output_dimensionality)
      .tolist()
  )
  return language_models.TextEmbedding(
      values=embeddings,
  )


def random_embeddings(
    texts: list[language_models.TextEmbeddingInput], output_dimensionality: int
) -> list[language_models.TextEmbedding]:
  return list(map(lambda x: _random_embedding(x, output_dimensionality), texts))


class PatchEmbeddingsModel:
  """This objects handles the patching of vertex ai embeddings models.

  This allows us to run tests without connecting to vertex ai. It will return
  a random embedding for any text.

  Context manager:
  ```
  with testing_utils.PatchEmbeddingsModel() as embeddings_model_patcher:
    model = models.VertexAIEmbeddings(model_name="test-model")
    response = model.embed_documents(["Example prompt"])
  ```

  Decorator:
  ```
  @testing_utils.PatchEmbeddingsModel()
  def test_example(self, embeddings_model_patcher):
    model = models.VertexAIEmbeddings(model_name="test-model")
    response = model.embed_documents(["Example prompt"])
  ```
  """

  def __init__(self):
    self.mock_embeddings_model = mock.MagicMock(
        spec=language_models.TextEmbeddingModel
    )

    self.mock_embeddings_model.get_embeddings.side_effect = random_embeddings
    self._from_pretrained_patcher = mock.patch(
        "vertexai.language_models.TextEmbeddingModel.from_pretrained",
        return_value=self.mock_embeddings_model,
    )

  def start(self):
    self.mock_from_pretrained = self._from_pretrained_patcher.start()

  def stop(self):
    self._from_pretrained_patcher.stop()

  def __call__(
      self, test_function: Callable[[Any], Any]
  ) -> Callable[[Any], Any]:
    def wrapped_test_function(*args, **kwargs):
      self.start()
      kwargs["embeddings_model_patcher"] = self
      test_function(*args, **kwargs)
      self.stop()

    return wrapped_test_function

  def __enter__(self) -> "PatchEmbeddingsModel":
    self.start()
    return self

  def __exit__(self, *args, **kwargs):
    self.stop()


class PatchGenerativeModel:
  """This objects handles the patching of vertex ai generative models.

  This allows us to run tests without connecting to vertex ai. It will
  always return the same response for any prompt. It can be used
  as either a decorator or a context manager:

  Context manager:
  ```
  with testing_utils.PatchGenerativeModel(
      response="Response text",
  ) as generative_model_patcher:
    model = generative_models.GenerativeModel(model_name="test-model")
    response = model.generate_content("Example prompt")
  ```

  Decorator:
  ```
  @testing_utils.PatchGenerativeModel(
      response="Response text",
  )
  def test_example(self, generative_model_patcher):
    model = generative_models.GenerativeModel(model_name="test-model")
    response = model.generate_content("Example prompt")
  ```
  """

  def __init__(self, response: str | GenerationResponse):
    self.mock_generative_model = mock.MagicMock(
        spec=generative_models.GenerativeModel
    )
    if isinstance(response, str):
      response = GenerationResponse.from_dict({
          "candidates": [
              {
                  "finish_reason": generative_models.FinishReason.STOP,
                  "content": {
                      "role": "model",
                      "parts": [{"text": response}],
                  }
              }
          ]
      })

    self.mock_generative_model.generate_content.return_value = response
    self.mock_generative_model.generate_content_async = mock.AsyncMock(
        return_value=response
    )

    self._model_init_patcher = mock.patch(
        "vertexai.generative_models.GenerativeModel",
        return_value=self.mock_generative_model,
    )

  def start(self):
    self.mock_init = self._model_init_patcher.start()

  def stop(self):
    self._model_init_patcher.stop()

  def __call__(
      self, test_function: Callable[[Any], Any]
  ) -> Callable[[Any], Any]:
    def wrapped_test_function(*args, **kwargs):
      self.start()
      kwargs["generative_model_patcher"] = self
      test_function(*args, **kwargs)
      self.stop()

    return wrapped_test_function

  def __enter__(self) -> "PatchGenerativeModel":
    self.start()
    return self

  def __exit__(self, *args, **kwargs):
    self.stop()


def values_are_equal(x: Any, y: Any) -> bool:
  """Recursively checks if the values of two iterables are equal."""
  if isinstance(x, str):
    # If the objects are strings, compare them directly
    return x == y

  if not hasattr(x, "__iter__") and not hasattr(y, "__iter__"):
    # If they are not iterables, compare them directly
    return x == y
  elif not hasattr(x, "__iter__") or not hasattr(y, "__iter__"):
    # Only one of them is an iterable, so they can't be equal
    return False

  # Otherwise, check they have the same length and compare them element-wise
  if len(x) != len(y):
    return False

  # If they are dicts, check they have the same keys and that the values are
  # equal.
  if isinstance(x, dict):
    for key in x.keys():
      if key not in y:
        return False
      if not values_are_equal(x[key], y[key]):
        return False
    return True

  # Finally, assume they are iterables and check they have the same elements.
  for x_i, y_i in zip(x, y):
    if not values_are_equal(x_i, y_i):
      return False

  return True


def vectorstore_instances_are_equal(
    ad_copy_vectorstore_1: ad_copy_generator.AdCopyVectorstore,
    ad_copy_vectorstore_2: ad_copy_generator.AdCopyVectorstore,
) -> bool:
  """Checks if two AdCopyVectorstores are equal.

  Args:
    ad_copy_vectorstore_1: The first AdCopyVectorstore to compare.
    ad_copy_vectorstore_2: The second AdCopyVectorstore to compare.

  Returns:
    True if the two AdCopyVectorstores are equal, False otherwise.
  """
  if not isinstance(ad_copy_vectorstore_1, ad_copy_generator.AdCopyVectorstore):
    return False

  if not isinstance(ad_copy_vectorstore_2, ad_copy_generator.AdCopyVectorstore):
    return False

  params_1 = {
      "embedding_model_name": ad_copy_vectorstore_1.embedding_model_name.value,
      "dimensionality": ad_copy_vectorstore_1.dimensionality,
      "embeddings_batch_size": ad_copy_vectorstore_1.embeddings_batch_size,
  }
  params_2 = {
      "embedding_model_name": ad_copy_vectorstore_2.embedding_model_name.value,
      "dimensionality": ad_copy_vectorstore_2.dimensionality,
      "embeddings_batch_size": ad_copy_vectorstore_2.embeddings_batch_size,
  }

  if not values_are_equal(params_1, params_2):
    return False

  try:
    pd.testing.assert_frame_equal(
        ad_copy_vectorstore_1.ad_exemplars,
        ad_copy_vectorstore_2.ad_exemplars,
        check_like=True,
    )
  except AssertionError:
    return False

  return True


def copycat_instances_are_equal(
    copycat_1: copycat.Copycat, copycat_2: copycat.Copycat
) -> bool:
  """Checks if two Copycat models are equal.

  Args:
    copycat_1: The first Copycat model to compare.
    copycat_2: The second Copycat model to compare.

  Returns:
    True if the two Copycat models are equal, False otherwise.
  """
  if not isinstance(copycat_1, copycat.Copycat):
    return False

  if not isinstance(copycat_2, copycat.Copycat):
    return False

  if not vectorstore_instances_are_equal(
      copycat_1.ad_copy_vectorstore, copycat_2.ad_copy_vectorstore
  ):
    return False

  if copycat_1.ad_format != copycat_2.ad_format:
    return False

  if copycat_1.style_guide != copycat_2.style_guide:
    return False

  return True



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap
from unittest.mock import MagicMock, patch

from absl.testing import absltest
from absl.testing import parameterized
from google.cloud import storage
from vertexai import generative_models
import pandas as pd

from copycat import ad_copy_generator
from copycat import style_guide
from copycat import testing_utils


class TestStyleGuideGenerator(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.test_response = generative_models.GenerationResponse.from_dict({
        "candidates": [
            {"content": {"parts": [{"text": "This is a test style guide"}]}}
        ]
    })
    self.embedding_model_patcher = testing_utils.PatchEmbeddingsModel()
    self.embedding_model_patcher.start()

  def tearDown(self):
    super().tearDown()
    self.embedding_model_patcher.stop()

  def test_get_all_files(self):
    mock_storage_client = MagicMock(spec=storage.Client)
    mock_bucket = MagicMock()
    mock_blob = MagicMock()
    mock_blob.name = "test_file.pdf"
    mock_blob.content_type = "application/pdf"

    mock_storage_client.bucket.return_value = mock_bucket
    mock_bucket.list_blobs.return_value = [mock_blob]

    with patch("google.cloud.storage.Client", return_value=mock_storage_client):
      generator = style_guide.StyleGuideGenerator()
      result = generator.get_all_files(bucket_name="test_bucket")

    expected_result = [{
        "uri": "gs://test_bucket/test_file.pdf",
        "mime_type": "application/pdf",
    }]
    self.assertEqual(result, expected_result)

    mock_storage_client.bucket.assert_called_once_with("test_bucket")
    mock_bucket.list_blobs.assert_called_once()

  def test_generate_style_guide(self):
    """Tests the style guide generation process with a mock model."""

    with testing_utils.PatchGenerativeModel(
        response=self.test_response
    ) as model_patcher:

      generator = style_guide.StyleGuideGenerator()
      generator.file_info = [
          {"uri": "gs://test_bucket/file1.pdf", "mime_type": "application/pdf"}
      ]

      response = generator.generate_style_guide(
          brand_name="Test Brand",
          additional_style_instructions="Write in a fun and friendly tone.",
          model_name="gemini-1.5-pro-002",
          temperature=0.8,
      )

      # Assertions
      self.assertLen(response.candidates, 1)  # Check for one candidate
      self.assertEqual(
          response.candidates[0].content.text,
          self.test_response.candidates[0].content.text,
      )  # Compare with the test response

      model_patcher.mock_generative_model.generate_content.assert_called_once()

  @parameterized.named_parameters([
      dict(
          testcase_name="No vectorstore, with additional instructions",
          with_ad_copy_vectorstore=False,
          additional_style_instructions="Write in a fun and friendly tone.",
          expected_text_prompt=textwrap.dedent("""\
            In these files is an ad report for Test Brand, containing their ads (headlines and descriptions) that they use on Google Search Ads for the corresponding keywords. Headlines and descriptions are lists, and Google constructs ads by combining those headlines and descriptions together into ads. Therefore the headlines and descriptions should be sufficiently varied that Google is able to try lots of different combinations in order to find what works best.

            Use the ad report to write a comprehensive style guide for this brand's ad copies that can serve as instruction for a copywriter to write new ad copies for Test Brand for new lists of keywords. Ensure that you capure strong phrases, slogans and brand names of Test Brand in the guide.

            Additionally, there could be other files included regarding the brand's style that you should consider in the style guide.

            Also incorporate the following style instructions into the style guide:

            Write in a fun and friendly tone."""),
      ),
      dict(
          testcase_name="With vectorstore, with additional instructions",
          with_ad_copy_vectorstore=True,
          additional_style_instructions="Write in a fun and friendly tone.",
          expected_text_prompt=textwrap.dedent("""\
            Below is an ad report for Test Brand, containing their ads (headlines and descriptions) that they use on Google Search Ads for the corresponding keywords. Headlines and descriptions are lists, and Google constructs ads by combining those headlines and descriptions together into ads. Therefore the headlines and descriptions should be sufficiently varied that Google is able to try lots of different combinations in order to find what works best.

            Use the ad report to write a comprehensive style guide for this brand's ad copies that can serve as instruction for a copywriter to write new ad copies for Test Brand for new lists of keywords. Ensure that you capure strong phrases, slogans and brand names of Test Brand in the guide.

            Additionally, there could be other files included regarding the brand's style that you should consider in the style guide.

            Also incorporate the following style instructions into the style guide:

            Write in a fun and friendly tone.

            Ad Report:

             {  "keywords":"keyword 1, keyword 2",  "headlines":[  "headline 1",  "headline 2"  ],  "descriptions":[  "description 1",  "description 2"  ] }

             {  "keywords":"keyword 3, keyword 4",  "headlines":[  "headline 3",  "headline 4"  ],  "descriptions":[  "description 3",  "description 4"  ] }
            
            """),
      ),
      dict(
          testcase_name="No vectorstore, no additional instructions",
          with_ad_copy_vectorstore=False,
          additional_style_instructions="",
          expected_text_prompt=textwrap.dedent(
              """\
              In these files is an ad report for Test Brand, containing their ads (headlines and descriptions) that they use on Google Search Ads for the corresponding keywords. Headlines and descriptions are lists, and Google constructs ads by combining those headlines and descriptions together into ads. Therefore the headlines and descriptions should be sufficiently varied that Google is able to try lots of different combinations in order to find what works best.

              Use the ad report to write a comprehensive style guide for this brand's ad copies that can serve as instruction for a copywriter to write new ad copies for Test Brand for new lists of keywords. Ensure that you capure strong phrases, slogans and brand names of Test Brand in the guide.

              Additionally, there could be other files included regarding the brand's style that you should consider in the style guide."""
          ),
      ),
      dict(
          testcase_name="With vectorstore, no additional instructions",
          with_ad_copy_vectorstore=True,
          additional_style_instructions="",
          expected_text_prompt=textwrap.dedent("""\
            Below is an ad report for Test Brand, containing their ads (headlines and descriptions) that they use on Google Search Ads for the corresponding keywords. Headlines and descriptions are lists, and Google constructs ads by combining those headlines and descriptions together into ads. Therefore the headlines and descriptions should be sufficiently varied that Google is able to try lots of different combinations in order to find what works best.

            Use the ad report to write a comprehensive style guide for this brand's ad copies that can serve as instruction for a copywriter to write new ad copies for Test Brand for new lists of keywords. Ensure that you capure strong phrases, slogans and brand names of Test Brand in the guide.

            Additionally, there could be other files included regarding the brand's style that you should consider in the style guide.

            Ad Report:

             {  "keywords":"keyword 1, keyword 2",  "headlines":[  "headline 1",  "headline 2"  ],  "descriptions":[  "description 1",  "description 2"  ] }

             {  "keywords":"keyword 3, keyword 4",  "headlines":[  "headline 3",  "headline 4"  ],  "descriptions":[  "description 3",  "description 4"  ] }
            
            """),
      ),
  ])
  def test_generate_style_guide_uses_expected_text_prompt(
      self,
      with_ad_copy_vectorstore,
      additional_style_instructions,
      expected_text_prompt,
  ):
    if with_ad_copy_vectorstore:
      training_data = pd.DataFrame.from_records([
          {
              "headlines": ["headline 1", "headline 2"],
              "descriptions": ["description 1", "description 2"],
              "keywords": "keyword 1, keyword 2",
          },
          {
              "headlines": ["headline 3", "headline 4"],
              "descriptions": ["description 3", "description 4"],
              "keywords": "keyword 3, keyword 4",
          },
      ])

      ad_copy_vectorstore = (
          ad_copy_generator.AdCopyVectorstore.create_from_pandas(
              training_data=training_data,
              embedding_model_name="text-embedding-004",
              dimensionality=256,
              max_initial_ads=100,
              max_exemplar_ads=10,
              affinity_preference=None,
              embeddings_batch_size=10,
              exemplar_selection_method="random",
          )
      )
    else:
      ad_copy_vectorstore = None

    with testing_utils.PatchGenerativeModel(
        response=self.test_response
    ) as model_patcher:
      generator = style_guide.StyleGuideGenerator()

      generator.generate_style_guide(
          brand_name="Test Brand",
          additional_style_instructions=additional_style_instructions,
          model_name="gemini-1.5-pro-002",
          temperature=0.8,
          ad_copy_vectorstore=ad_copy_vectorstore,
      )

      text_prompt = (
          model_patcher.mock_generative_model.generate_content.call_args.kwargs[
              "contents"
          ][0]
          .parts[-1]
          .text
      )
      self.assertEqual(text_prompt, expected_text_prompt)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import re

import pydantic
from sklearn.metrics import pairwise

from copycat import ad_copy_generator
from copycat import google_ads


GoogleAd = google_ads.GoogleAd
GoogleAdFormat = google_ads.GoogleAdFormat


class EvaluationResults(pydantic.BaseModel):
  """The metrics used to evaluate the generated ad."""

  headlines_are_memorised: bool | None
  descriptions_are_memorised: bool | None
  errors: list[str]
  warnings: list[str]
  style_similarity: float | None
  keyword_similarity: float | None


def _normalized_cosine_similarity(
    embedding_1: list[float], embedding_2: list[float]
) -> float:
  """Calculate the cosine similarity normalized to a value between 0 and 1."""
  similarity = float(
      pairwise.cosine_similarity([embedding_1], [embedding_2])[0][0]
  )
  return min(max((1.0 + similarity) / 2.0, 0.0), 1.0)


@dataclasses.dataclass
class AdCopyEvaluator:
  """Evaluates the ad copy.

  Attributes:
    ad_format: The ad format.
    ad_copy_vectorstore: The vectorstore containing the training ads, if
      provided. Defaults to None.
    training_headlines: The headlines in the training data.
    training_descriptions: The descriptions in the training data.
  """

  ad_format: GoogleAdFormat
  ad_copy_vectorstore: ad_copy_generator.AdCopyVectorstore | None = None

  @property
  def training_headlines(self) -> set[str]:
    if self.ad_copy_vectorstore is None:
      return set()
    return self.ad_copy_vectorstore.unique_headlines

  @property
  def training_descriptions(self) -> set[str]:
    if self.ad_copy_vectorstore is None:
      return set()
    return self.ad_copy_vectorstore.unique_descriptions

  def has_valid_number_of_headlines(self, ad_copy: GoogleAd) -> bool:
    """Returns true if the number of headlines is valid.

    The number of headlines is valid if it is less than or equal to the maximum
    number of headlines for the ad format, and greater than or equal the minimum
    number of headlines for the ad format.

    Args:
      ad_copy: The ad copy to evaluate.
    """
    return (ad_copy.headline_count <= self.ad_format.max_headlines) and (
        ad_copy.headline_count >= self.ad_format.min_headlines
    )

  def has_valid_number_of_descriptions(self, ad_copy: GoogleAd) -> bool:
    """Returns true if the number of descriptions is valid.

    The number of descriptions is valid if it is less than or equal to the
    maximum number of descriptions for the ad format, and greater than or equal
    to the minimum number of descriptions for the ad format.

    Args:
      ad_copy: The ad copy to evaluate.
    """
    return (ad_copy.description_count <= self.ad_format.max_descriptions) and (
        ad_copy.description_count >= self.ad_format.min_descriptions
    )

  def has_valid_headline_lengths(self, ad_copy: GoogleAd) -> bool:
    """Returns true if the lengths of the headlines are valid.

    The lengths of the headlines are valid if they are all less than or equal to
    the maximum length for a headline.

    Args:
      ad_copy: The ad copy to evaluate.
    """
    return all(
        len(google_ads.parse_google_ads_special_variables(headline))
        <= self.ad_format.max_headline_length
        for headline in ad_copy.headlines
    )

  def has_valid_description_lengths(self, ad_copy: GoogleAd) -> bool:
    """Returns true if the lengths of the descriptions are valid.

    The lengths of the descriptions are valid if they are all less than or equal
    to the maximum length for a description.

    Args:
      ad_copy: The ad copy to evaluate.
    """
    return all(
        len(google_ads.parse_google_ads_special_variables(description))
        <= self.ad_format.max_description_length
        for description in ad_copy.descriptions
    )

  def has_unique_headlines(self, ad_copy: GoogleAd) -> bool:
    """Returns true if there are no duplicate headlines.

    Args:
      ad_copy: The ad copy to evaluate.
    """
    return len(set(ad_copy.headlines)) == len(ad_copy.headlines)

  def has_unique_descriptions(self, ad_copy: GoogleAd) -> bool:
    """Returns true if there are no duplicate descriptions.

    Args:
      ad_copy: The ad copy to evaluate.
    """
    return len(set(ad_copy.descriptions)) == len(ad_copy.descriptions)

  def is_valid(self, ad_copy: GoogleAd) -> bool:
    """Returns true if the ad copy is valid.

    This checks that the number of headlines and descriptions is within the
    limits for the ad format, the headlines and descriptions are unique, and
    that the length of each headline and description is within the limits.

    Args:
      ad_copy: The ad copy to evaluate.
    """

    return (
        self.has_valid_number_of_headlines(ad_copy)
        and self.has_valid_number_of_descriptions(ad_copy)
        and self.has_valid_headline_lengths(ad_copy)
        and self.has_valid_description_lengths(ad_copy)
        and self.has_unique_headlines(ad_copy)
        and self.has_unique_descriptions(ad_copy)
    )

  def has_unfillable_google_ads_special_variables(
      self, ad_copy: GoogleAd
  ) -> bool:
    """Returns true if the ad contains special variables that cannot be filled.

    Special variables are things like Dynamic Keyword Insertion (DKI) and
    Customizers, and they are always filled with their default if possible
    before being sent to Copycat.

    If a special variable cant be filled because it doesn't have a default,
    then it will be left as it is. This function checks for this case,
    because these ads should not be sent to Copycat, but will need to be
    handled manually by the user.

    Args:
      ad_copy: The ad copy to evaluate.
    """
    for headline in ad_copy.headlines:
      cleaned_headline = google_ads.parse_google_ads_special_variables(headline)
      unfilled_variables = re.findall(r"({.*?})", cleaned_headline)
      if unfilled_variables:
        return True

    for description in ad_copy.descriptions:
      cleaned_description = google_ads.parse_google_ads_special_variables(
          description
      )
      unfilled_variables = re.findall(r"({.*?})", cleaned_description)
      if unfilled_variables:
        return True

    return False

  def is_complete(self, ad_copy: GoogleAd) -> bool:
    """Returns true if the ad copy is complete.

    A complete ad copy contains the maximum number of headlines and
    descriptions.

    Args:
      ad_copy: The ad copy to evaluate.
    """
    complete_headlines = len(ad_copy.headlines) == self.ad_format.max_headlines
    complete_descriptions = (
        len(ad_copy.descriptions) == self.ad_format.max_descriptions
    )
    return complete_headlines and complete_descriptions

  def is_empty(self, ad_copy: GoogleAd) -> bool:
    """Returns true if the ad copy is empty."""
    return not ad_copy.headlines and not ad_copy.descriptions

  def is_underpopulated(self, ad_copy: GoogleAd) -> bool:
    """Returns true if the ad copy is underpopulated.

    This means the ad copy has fewer than the minimum number of headlines or
    descriptions.

    Args:
      ad_copy: The ad copy to evaluate.
    """
    return (len(ad_copy.headlines) < self.ad_format.min_headlines) or (
        len(ad_copy.descriptions) < self.ad_format.min_descriptions
    )

  def headlines_are_memorised(self, ad_copy: GoogleAd) -> bool:
    """Returns true if all the headlines exist in the training data."""
    if not ad_copy.headlines:
      # There are no headlines, so they cannot be memorised.
      return False

    return not (set(ad_copy.headlines) - self.training_headlines)

  def descriptions_are_memorised(self, ad_copy: GoogleAd) -> bool:
    """Returns true if all the descriptions exist in the training data."""
    if not ad_copy.descriptions:
      # There are no descriptions, so they cannot be memorised.
      return False

    return not (set(ad_copy.descriptions) - self.training_descriptions)

  def calculate_similarity_metrics_batch(
      self,
      *,
      ad_copies: list[GoogleAd],
      keywords: list[str],
  ) -> list[dict[str, float]]:
    """Evaluates the generated ad copies against the training data and keywords.

    This calculates two metrics:
      - The style similarity, which is how similar the style of the generated ad
        is to the style of the training ads. It is calculated as the similarity
        of the generated ad to the most similar training ad.
      - The keyword similarity, which is how similar the generated ad is to the
        keywords.

    The similarity in both cases is calculated using the cosine similarity, and
    normalising it between 0 and 1, so 0 is the least similar and 1 is the most
    similar.

    Args:
      ad_copies: The generated ads.
      keywords: The keywords used to generate the ads.

    Returns:
      A list of dicts containing the style_similarity and keyword_similarity for
      each ad. The dict will be empty if the vectorstore is not provided or
      if the ad has no headlines and descriptions.
    """
    if self.ad_copy_vectorstore is None:
      return [dict()] * len(ad_copies)

    keywords_embeddings = self.ad_copy_vectorstore.embed_queries(keywords)
    ad_embeddings = self.ad_copy_vectorstore.embed_documents(
        list(map(str, ad_copies))
    )

    _, similar_training_ads_embeddings = (
        self.ad_copy_vectorstore.get_relevant_ads_and_embeddings_from_embeddings(
            ad_embeddings, k=1
        )
    )
    most_similar_training_ads_embeddings = [
        similar_embeddings[0]
        for similar_embeddings in similar_training_ads_embeddings
    ]

    similarity_metrics = []
    for (
        keywords_embedding,
        ad_embedding,
        ad_copy,
        most_similar_training_ad_embeddings,
    ) in zip(
        keywords_embeddings,
        ad_embeddings,
        ad_copies,
        most_similar_training_ads_embeddings,
    ):
      if self.is_empty(ad_copy):
        similarity_metrics.append(dict())
        continue

      keyword_similarity = _normalized_cosine_similarity(
          keywords_embedding, ad_embedding
      )

      style_similarity = _normalized_cosine_similarity(
          most_similar_training_ad_embeddings, ad_embedding
      )

      similarity_metrics.append({
          "style_similarity": style_similarity,
          "keyword_similarity": keyword_similarity,
      })

    return similarity_metrics

  def _evaluate_simple_metrics(
      self,
      ad_copy: GoogleAd,
      *,
      allow_memorised_headlines: bool = False,
      allow_memorised_descriptions: bool = False,
  ) -> EvaluationResults:
    """Evaluates the generated ad copy.

    This checks that the number of headlines and descriptions is within the
    limits for the ad format, the headlines and descriptions are unique, and
    that the length of each headline and description is within the limits.

    Args:
      ad_copy: The ad copy to evaluate.
      allow_memorised_headlines: Whether to allow the headlines to be memorised.
      allow_memorised_descriptions: Whether to allow the descriptions to be
        memorised.

    Returns:
      The evaluation results for the ad copy.
    """
    errors = []
    warnings = []

    if not self.has_valid_number_of_headlines(ad_copy):
      errors.append("Invalid number of headlines for the ad format.")
    if not self.has_valid_number_of_descriptions(ad_copy):
      errors.append("Invalid number of descriptions for the ad format.")
    if not self.has_valid_headline_lengths(ad_copy):
      errors.append("At least one headline too long for the ad format.")
    if not self.has_valid_description_lengths(ad_copy):
      errors.append("At least one description too long for the ad format.")
    if not self.has_unique_headlines(ad_copy):
      errors.append("Duplicate headlines found.")
    if not self.has_unique_descriptions(ad_copy):
      errors.append("Duplicate descriptions found.")

    headlines_are_memorised = self.headlines_are_memorised(ad_copy)
    descriptions_are_memorised = self.descriptions_are_memorised(ad_copy)

    if headlines_are_memorised:
      if allow_memorised_headlines:
        warnings.append("All headlines are memorised from the training data.")
      else:
        errors.append("All headlines are memorised from the training data.")
    if descriptions_are_memorised:
      if allow_memorised_descriptions:
        warnings.append(
            "All descriptions are memorised from the training data."
        )
      else:
        errors.append("All descriptions are memorised from the training data.")

    return EvaluationResults(
        errors=errors,
        warnings=warnings,
        headlines_are_memorised=headlines_are_memorised,
        descriptions_are_memorised=descriptions_are_memorised,
        style_similarity=None,
        keyword_similarity=None,
    )

  def evaluate_batch(
      self,
      ad_copies: list[GoogleAd],
      *,
      allow_memorised_headlines: bool = False,
      allow_memorised_descriptions: bool = False,
      keywords: list[str] | None = None,
  ) -> list[EvaluationResults]:
    """Evaluates the generated ad copies.

    Args:
      ad_copies: The generated ads.
      allow_memorised_headlines: Whether to allow the headlines to be memorised.
      allow_memorised_descriptions: Whether to allow the descriptions to be
        memorised.
      keywords: The list of keywords used to generate the ads. Only required for
        the style and keyword similarity metrics. If not provided, these metrics
        will be None.

    Returns:
      The list of evaluation results for each ad.
    """

    results = [
        self._evaluate_simple_metrics(
            ad_copy=ad_copy,
            allow_memorised_headlines=allow_memorised_headlines,
            allow_memorised_descriptions=allow_memorised_descriptions,
        )
        for ad_copy in ad_copies
    ]

    if keywords is None:
      return results

    similarity_metrics = self.calculate_similarity_metrics_batch(
        ad_copies=ad_copies, keywords=keywords
    )

    results = [
        results_i.model_copy(
            update=dict(
                style_similarity=similarity_metrics_i.get("style_similarity"),
                keyword_similarity=similarity_metrics_i.get(
                    "keyword_similarity"
                ),
            )
        )
        for results_i, similarity_metrics_i in zip(results, similarity_metrics)
    ]
    return results



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An object for handling the sub pages of the Copycat UI."""

from typing import Callable

import mesop as me

from copycat.ui import components
from copycat.ui import states
from copycat.ui import styles


def navigate_to_subpage(event: me.ClickEvent):
  state = me.state(states.AppState)
  state.highlighted_url = event.key


class SubPages(dict):
  """An object for handling the sub pages of the Copycat UI.

  Stores functions for rendering the pages, and the navigation button text and
  icon, in a dictionary. The keys of the dictionary are the urls of the pages.
  """

  def add_page(
      self, page_func: Callable[[], None], *, nav_text: str, nav_icon: str
  ) -> None:
    """Adds a page to the sub pages.

    Args:
      page_func: The function to call to render the page.
      nav_text: The text to display in the navigation button.
      nav_icon: The icon to display in the navigation button.
    """
    url = "/" + getattr(page_func, "__name__", repr(page_func))
    self[url] = dict(
        page_func=page_func,
        nav_text=nav_text,
        nav_icon=nav_icon,
    )

  @me.content_component
  def navigation_button(self, url: str):
    """Renders the navigation button for a sub page.

    Args:
      url: The url of the sub page.
    """
    state = me.state(states.AppState)

    style_params = dict(
        padding=me.Padding.symmetric(vertical=30, horizontal=25),
    )
    if state.highlighted_url == url:
      style_params["background"] = me.theme_var("primary")
      style_params["color"] = me.theme_var("surface-container")

    with me.content_button(
        key=url, on_click=navigate_to_subpage, style=me.Style(**style_params)
    ):
      me.icon(self[url]["nav_icon"])
      me.text(self[url]["nav_text"])

  def render(self, **kwargs):
    """Renders the nav bar and the body of the active sub page.

    Args:
      **kwargs: Any additional keyword arguments to pass to the outer box
        containing the nav bar and the body.
    """
    state = me.state(states.AppState)

    with components.column(**kwargs):
      # Nav Bar
      with components.header_bar(
          border=me.Border(bottom=styles.DEFAULT_BORDER_STYLE)
      ):
        with components.header_section():
          for i, url in enumerate(self.keys()):
            with self.navigation_button(url):
              pass
            if i < (len(self.keys()) - 1):
              me.icon(
                  "arrow_forward_ios",
                  style=me.Style(color=me.theme_var("primary")),
              )

      # Body
      with me.box(
          style=me.Style(
              padding=me.Padding.all(24),
              overflow_y="auto",
              display="flex",
              height="100%",
              width="100%",
              gap=5,
          )
      ):
        if state.highlighted_url:
          self[state.highlighted_url]["page_func"]()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Re-usable styles for the Copycat UI."""

import mesop as me


DEFAULT_ROW_AND_COLUMN_GAP = 16
HEADER_SECTION_STYLE = dict(display="flex", gap=5, align_items="center")
ROUNDED_BOX_SECTION_STYLE = dict(
    border_radius=16,
    padding=me.Padding.all(24),
    border=me.Border.all(
        me.BorderSide(
            width=1,
            style="solid",
            color=me.theme_var("outline-variant"),
        )
    ),
)
ROUNDED_BOX_SECTION_HEADER_TYPE = "headline-5"

DEFAULT_BORDER_STYLE = me.BorderSide(
    width=1,
    style="solid",
    color=me.theme_var("outline-variant"),
)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A mock gspread client for testing."""

from typing import Any

import gspread
import mock

MOCK_SHEETS_URL_PREFIX = "https://mock.sheets.com/spreadsheet/"


def _empty_formatting(n_rows: int, n_cols: int):
  return [[{} for _ in range(n_cols)] for _ in range(n_rows)]


def _empty_data(n_rows: int, n_cols: int):
  return [["" for _ in range(n_cols)] for _ in range(n_rows)]


class MockWorksheet:
  """A mock gspread worksheet for testing.

  This mock is not comprehensive, it just covers the functionality that is used
  by the Copycat library.
  """

  def __init__(
      self,
      title: str,
      data: list[list[Any]],
      formatting: list[list[dict[str, Any]]],
      frozen_col_count: int = 0,
      frozen_row_count: int = 0,
  ):
    """Initialises the mock worksheet."""
    self.title = title
    self._data = data
    self._formatting = formatting
    self._frozen_col_count = frozen_col_count
    self._frozen_row_count = frozen_row_count

  @property
  def frozen_col_count(self) -> int:
    return self._frozen_col_count

  @property
  def frozen_row_count(self) -> int:
    return self._frozen_row_count

  @property
  def row_count(self) -> int:
    return len(self._data)

  @property
  def col_count(self) -> int:
    return len(self._data[0])

  def get_all_records(self, *args, **kwargs) -> list[dict[str, Any]]:
    """Returns all the records in the worksheet."""
    columns = list(filter(None, self._data[0]))
    rows = list(filter(None, [row[: len(columns)] for row in self._data[1:]]))

    # Find the index of the first non-empty row (counting backwards).
    for i, row in enumerate(rows[::-1]):
      # Start at the end of the worksheet and work backwards.
      if list(filter(None, row)):
        # Stop counting at the first non-empty row.
        break
    else:
      # If there are no non-empty rows, set i to the length of the rows list.
      i = len(rows)

    # Remove the empty rows from the end of the worksheet.
    last_idx = len(rows) - i
    rows = rows[:last_idx]

    return [dict(zip(columns, row)) for row in rows]

  def clear(self) -> None:
    """Clears the worksheet."""
    self._data = _empty_data(self.row_count, self.col_count)
    self._formatting = _empty_formatting(self.row_count, self.col_count)

  def freeze(self, rows: int | None = None, cols: int | None = None) -> None:
    """Freezes the given number of rows and columns."""
    if rows and rows > self.row_count:
      raise ValueError(
          "Cannot freeze more rows than there are in the worksheet"
      )
    if cols and cols > self.col_count:
      raise ValueError(
          "Cannot freeze more columns than there are in the worksheet"
      )

    if rows is not None:
      self._frozen_row_count = rows
    if cols is not None:
      self._frozen_col_count = cols

  def format(self, ranges: str | list[str], format: dict[str, Any]) -> None:
    """Formats the given range with the given format."""
    if isinstance(ranges, str):
      ranges = [ranges]

    for range_str in ranges:
      grid_range = gspread.utils.a1_range_to_grid_range(range_str)
      default_grid_range = {
          "startRowIndex": 0,
          "endRowIndex": self.row_count,
          "startColumnIndex": 0,
          "endColumnIndex": self.col_count,
      }
      grid_range = default_grid_range | grid_range

      start_row = grid_range["startRowIndex"]
      end_row = grid_range["endRowIndex"]
      start_col = grid_range["startColumnIndex"]
      end_col = grid_range["endColumnIndex"]

      for row in range(start_row, end_row):
        for col in range(start_col, end_col):
          self._formatting[row][col].update(format)

  def add_rows(self, rows: int) -> None:
    """Adds empty rows to the worksheet."""
    self._data.extend(_empty_data(rows, self.col_count))
    self._formatting.extend(_empty_formatting(rows, self.col_count))

  def delete_rows(self, start_index: int, end_index: int | None = None) -> None:
    """Deletes rows from the worksheet."""
    if end_index is None:
      end_index = start_index

    self._data = self._data[:start_index] + self._data[end_index + 1 :]
    self._formatting = (
        self._formatting[:start_index] + self._formatting[end_index + 1 :]
    )

  def add_cols(self, n_cols: int) -> None:
    """Adds n_cols to the worksheet."""
    new_cols_data = _empty_data(self.row_count, n_cols)
    new_cols_formatting = _empty_formatting(self.row_count, n_cols)

    for i in range(self.row_count):
      self._data[i].extend(new_cols_data[i])
      self._formatting[i].extend(new_cols_formatting[i])

  def delete_columns(
      self, start_index: int, end_index: int | None = None
  ) -> None:
    """Deletes columns from the worksheet."""
    if end_index is None:
      end_index = start_index

    for i in range(self.row_count):
      self._data[i] = (
          self._data[i][:start_index] + self._data[i][end_index + 1 :]
      )
      self._formatting[i] = (
          self._formatting[i][:start_index]
          + self._formatting[i][end_index + 1 :]
      )

  def update(
      self,
      values: list[list[Any]],
      range_name: str | None = None,
      *args,
      **kwargs
  ) -> None:
    """Updates the data in the worksheet."""
    if range_name is None:
      start_row = 0
      start_col = 0
      target_rows = self.row_count
      target_cols = self.col_count
    else:
      grid_range = gspread.utils.a1_range_to_grid_range(range_name)
      default_grid_range = {
          "startRowIndex": 0,
          "endRowIndex": self.row_count,
          "startColumnIndex": 0,
          "endColumnIndex": self.col_count,
      }
      grid_range = default_grid_range | grid_range

      grid_range["endRowIndex"] = min(grid_range["endRowIndex"], self.row_count)
      grid_range["endColumnIndex"] = min(
          grid_range["endColumnIndex"], self.col_count
      )
      target_rows = grid_range["endRowIndex"] - grid_range["startRowIndex"]
      target_cols = (
          grid_range["endColumnIndex"] - grid_range["startColumnIndex"]
      )

      start_row = grid_range["startRowIndex"]
      start_col = grid_range["startColumnIndex"]

    if target_rows < len(values):
      raise ValueError(
          "Cannot update more rows than there are in the worksheet / range"
      )
    if target_cols < len(values[0]):
      raise ValueError(
          "Cannot update more columns than there are in the worksheet / range"
      )

    for i, row in enumerate(values):
      for j, cell in enumerate(row):
        self._data[start_row + i][start_col + j] = cell

  def batch_update(
      self, batches: list[dict[str, str | list[list[Any]]]]
  ) -> None:
    """Updates the data in the worksheet in batches."""
    for batch in batches:
      self.update(batch["values"], range_name=batch["range"])

  def insert_row(self, values: list[Any], index: int) -> None:
    """Inserts a row at the given index.

    Args:
      values: The values to insert into the row.
      index: The index to insert the row at. The index starts at 1, so the first
        row is row 1.
    """
    if len(values) > self.col_count:
      self.add_cols(len(values) - self.col_count)
    if len(values) < self.col_count:
      values.extend(_empty_data(1, self.col_count - len(values))[0])

    self._data.insert(index - 1, values)
    self._formatting.insert(index, _empty_formatting(1, self.col_count)[0])

  def row_values(self, index: int) -> list[Any]:
    """Returns the values in the given row.

    Args:
      index: The index of the row to return. The index starts at 1, so the first
        row is row 1.
    """
    return self._data[index - 1]


class MockSpreadsheet:
  """A mock gspread spreadsheet for testing.

  This mock is not comprehensive, it just covers the functionality that is used
  by the Copycat library.
  """

  def __init__(self, title: str):
    self.title = title
    self.url = MOCK_SHEETS_URL_PREFIX + title
    self._worksheets = {}

    self.add_worksheet("Sheet1")

  def worksheet(self, title: str) -> MockWorksheet:
    """Returns the worksheet with the given name."""
    return self._worksheets[title]

  def worksheets(self) -> list[MockWorksheet]:
    """Returns the worksheets in the spreadsheet."""
    return list(self._worksheets.values())

  def add_worksheet(
      self, title: str, rows: int = 1000, cols: int = 26
  ) -> MockWorksheet:
    """Adds an empty worksheet to the spreadsheet."""
    data = _empty_data(rows, cols)
    formatting = _empty_formatting(rows, cols)
    self._worksheets[title] = MockWorksheet(
        title, data=data, formatting=formatting
    )
    return self.worksheet(title)

  def del_worksheet(self, worksheet: MockWorksheet) -> None:
    """Deletes the given worksheet from the spreadsheet."""
    self._worksheets.pop(worksheet.title)


class MockGspreadClient:
  """A mock gspread client for testing.

  This mock is not comprehensive, it just covers the functionality that is used
  by the Copycat library.
  """

  def __init__(self):
    self._spreadsheets = {}

  def create(self, title: str) -> MockSpreadsheet:
    """Creates a mock spreadsheet."""
    new_sheet = MockSpreadsheet(title)
    self._spreadsheets[new_sheet.url] = new_sheet
    return new_sheet

  def open_by_url(self, url: str) -> MockSpreadsheet:
    """Opens a mock spreadsheet by URL."""
    if url not in self._spreadsheets:
      raise gspread.SpreadsheetNotFound("Invalid URL")

    return self._spreadsheets[url]


class PatchGspread:
  """This objects handles the patching of gspread.

  This allows us to run tests without connecting to google sheets. It will
  return a mock gspread client which can handle reading and writing data
  locally.

  Context manager:
  ```
  with testing_utils.PatchGspread() as gspread_patcher:
    creds, _ = google.auth.default()
    client = gspread.authorize(creds)
  ```

  Manually calling start and stop:
  ```
  patcher = PatchGspread()
  patcher.start()
  creds, _ = google.auth.default()
  client = gspread.authorize(creds)
  patcher.stop()
  ```
  """

  def __init__(self):
    self.mock_client = MockGspreadClient()
    self._gspread_authorize_patcher = mock.patch(
        "gspread.authorize",
        return_value=self.mock_client,
    )

  def start(self):
    self.mock_gspread_authorize = self._gspread_authorize_patcher.start()

  def stop(self):
    self._gspread_authorize_patcher.stop()

  def __enter__(self) -> "PatchGspread":
    self.start()
    return self

  def __exit__(self, *args, **kwargs):
    self.stop()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for loading and saving data to Google Sheets."""

import logging
import pathlib
import time
from typing import Any
import google.auth.credentials
import gspread
import pandas as pd


TEMPLATE_EXISTING_ADS_WORKSHEET_NAME = "Training Ads"
TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME = "New Keywords"
TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME = "Extra Instructions for New Ads"


HEADING_FORMAT = {
    "backgroundColor": {"red": 0.3, "green": 0.3, "blue": 0.3},
    "horizontalAlignment": "CENTER",
    "textFormat": {
        "bold": True,
        "foregroundColor": {"red": 1.0, "green": 1.0, "blue": 1.0},
    },
}


GOOGLE_AUTH_CREDENTIALS: google.auth.credentials.Credentials | None = None


def set_google_auth_credentials(
    credentials: google.auth.credentials.Credentials,
) -> None:
  """Sets the Google auth credentials.

  These credentials are used to authorize the Google Sheets client. You can get
  the credentials by calling google.auth.default().
  """
  global GOOGLE_AUTH_CREDENTIALS
  GOOGLE_AUTH_CREDENTIALS = credentials


def get_gspread_client() -> gspread.Client:
  """Creates a Google Sheets client."""
  if GOOGLE_AUTH_CREDENTIALS is None:
    raise ValueError(
        "Google auth credentials are not set. First call "
        "set_google_auth_credentials()"
    )
  return gspread.authorize(GOOGLE_AUTH_CREDENTIALS)


class GoogleSheet:
  """An object for reading and writing data to Google Sheets.

  This class provides a convenient way to access and manipulate data in Google
  Sheets. It allows you to read and write pandas dataframes to/from Google
  Sheets.

  Attributes:
    spreadsheet: The Google Sheets spreadsheet.
    url: The url of the spreadsheet.
    title: The title of the spreadsheet.
  """

  def __init__(self, spreadsheet: gspread.Spreadsheet):
    """Initializes the GoogleSheet object.

    Args:
      spreadsheet: The Google Sheets spreadsheet.
    """
    self.spreadsheet = spreadsheet

  @property
  def url(self) -> str:
    return self.spreadsheet.url

  @property
  def title(self) -> str:
    return self.spreadsheet.title

  def __str__(self):
    return "\n".join([
        "GoogleSheet:",
        f"  URL: {self.url}",
        f"  Name: {self.title}",
        (
            "  Worksheet Names:"
            f" {[worksheet.title for worksheet in self.spreadsheet.worksheets()]}"
        ),
    ])

  @classmethod
  def new(cls, spreadsheet_name: str) -> "GoogleSheet":
    """Creates a Google Sheet DataFrame with a new spreadsheet.

    Args:
      spreadsheet_name: The name of the spreadsheet to create.

    Returns:
      A GoogleSheet object referencing the new spreadsheet.
    """
    spreadsheet = get_gspread_client().create(spreadsheet_name)
    return cls(spreadsheet)

  @classmethod
  def load(cls, url: str) -> "GoogleSheet":
    """Loads a Google Sheet DataFrame from a URL.

    Args:
      url: The URL of the spreadsheet to load.

    Returns:
      A GoogleSheet object referencing the spreadsheet.
    """
    spreadsheet = get_gspread_client().open_by_url(url)
    return cls(spreadsheet)

  def __contains__(self, worksheet_title: str) -> bool:
    """Checks if a worksheet with the given title exists in the spreadsheet.

    Args:
      worksheet_title: The title of the worksheet to check.

    Returns:
      True if the worksheet exists, False otherwise.
    """
    for worksheet in self.spreadsheet.worksheets():
      if worksheet.title == worksheet_title:
        return True
    return False

  def __getitem__(self, worksheet_title: str) -> pd.DataFrame:
    """Gets the data from a worksheet with the given title.

    The column headings are assumed to be the first row of the worksheet.
    The row index is assumed to be all of the frozen columns in the worksheet.

    Args:
      worksheet_title: The title of the worksheet to get data from.

    Returns:
      A pandas DataFrame containing the data from the worksheet.
    """
    worksheet = self.spreadsheet.worksheet(worksheet_title)
    records = worksheet.get_all_records(value_render_option="UNFORMATTED_VALUE")

    if records:
      data = pd.DataFrame(records)
    else:
      column_names = list(filter(None, worksheet.row_values(1)))
      data = pd.DataFrame(columns=column_names)

    index_cols = data.columns.values[: worksheet.frozen_col_count].tolist()
    if index_cols:
      data = data.set_index(index_cols)

    return data

  def _parse_data(
      self, data: pd.DataFrame
  ) -> tuple[list[list[Any]], list[str], int]:
    """Parses the data into a format suitable for writing to Google Sheets.

    Args:
      data: A pandas DataFrame containing the data to parse.

    Returns:
      A tuple containing the data to write, the column names, and the number of
      index columns.
    """
    n_index_cols = data.index.nlevels
    data_reset = data.reset_index()
    column_names = data_reset.columns.tolist()
    row_values = data_reset.values.tolist()

    if not row_values:
      # If there is no data, then just have a single empty row
      row_values = [[""] * len(column_names)]

    data_to_write = [column_names] + row_values
    return data_to_write, column_names, n_index_cols

  def _overwrite_worksheet(
      self, worksheet_title: str, data: list[list[Any]], n_index_cols: int
  ) -> None:
    """Overwrites the data in a worksheet with the given title.

    If the worksheet does not exist, then it is created.

    Args:
      worksheet_title: The title of the worksheet to overwrite.
      data: A list of lists containing the data to write.
      n_index_cols: The number of columns to freeze as indexes.
    """
    if worksheet_title in self:
      worksheet = self.spreadsheet.worksheet(worksheet_title)
      worksheet.clear()
    else:
      worksheet = self.spreadsheet.add_worksheet(
          title=worksheet_title,
          rows=len(data),
          cols=len(data[0]),
      )
    self._update_size_of_worksheet(worksheet_title, data)
    worksheet.update(data, raw=False)
    self._update_worksheet_formatting(worksheet_title, n_index_cols)

  def _update_worksheet_formatting(
      self, worksheet_title: str, n_index_cols: int
  ) -> None:
    """Updates the formatting of the worksheet."""
    worksheet = self.spreadsheet.worksheet(worksheet_title)
    worksheet.freeze(rows=1, cols=n_index_cols)
    worksheet.format("1:1", HEADING_FORMAT)

  def _update_size_of_worksheet(
      self, worksheet_title: str, data: list[list[Any]]
  ) -> None:
    """Updates the size of the worksheet to fit the data."""
    worksheet = self.spreadsheet.worksheet(worksheet_title)
    target_rows = len(data)
    target_cols = len(data[0])
    if worksheet.row_count < target_rows:
      worksheet.add_rows(target_rows - worksheet.row_count)
    elif worksheet.row_count > target_rows:
      worksheet.delete_rows(1, worksheet.row_count - target_rows)

    if worksheet.col_count < target_cols:
      worksheet.add_cols(target_cols - worksheet.col_count)
    elif worksheet.col_count > target_cols:
      worksheet.delete_columns(1, worksheet.col_count - target_cols)

  def _construct_update_batches(
      self,
      data_to_write: list[list[Any]],
      existing_data: list[list[Any]],
  ) -> list[dict[str, Any]]:
    """Constructs update batches for Google Sheets.

    The batches are the rows where the data is different.

    Args:
      data_to_write: A list of lists containing the data to write.
      existing_data: A list of lists containing the existing data.

    Returns:
      A list of update batches.
    """
    update_batches = []
    last_different_row_index = -1
    for row_index, (new_row, old_row) in enumerate(
        zip(data_to_write, existing_data), start=1
    ):
      if new_row != old_row:
        if last_different_row_index == (row_index - 1):
          update_batches[-1]["end_row"] = row_index
          update_batches[-1]["values"].append(new_row)
        else:
          update_batches.append({
              "start_row": row_index,
              "end_row": row_index,
              "values": [new_row],
          })
        last_different_row_index = row_index

    if len(data_to_write) > len(existing_data):
      update_batches.append({
          "start_row": len(existing_data) + 1,
          "end_row": len(data_to_write),
          "values": data_to_write[len(existing_data) :],
      })

    update_batches = [
        {
            "range": (
                gspread.utils.rowcol_to_a1(batch["start_row"], 1)
                + ":"
                + gspread.utils.rowcol_to_a1(
                    batch["end_row"], len(data_to_write[0])
                )
            ),
            "values": batch["values"],
        }
        for batch in update_batches
    ]
    return update_batches

  def __setitem__(self, worksheet_title: str, data: pd.DataFrame) -> None:
    """Sets the data in a worksheet with the given title.

    The first row of the data is set with the column headings.
    The indexes are set as the first column(s) and are frozen.
    If the worksheet already exists and column names are unchanged, then only
    the rows that have changed are updated. If the worksheet already exists and
    the column names are changed, then the entire worksheet is overwritten.

    Args:
      worksheet_title: The title of the worksheet to set data in.
      data: A pandas DataFrame containing the data to set.
    """
    data_to_write, column_names, n_index_cols = self._parse_data(data)

    # If the worksheet does not exist, then just write all the data
    if worksheet_title not in self:
      self._overwrite_worksheet(worksheet_title, data_to_write, n_index_cols)
      return

    existing_data, existing_columns, _ = self._parse_data(self[worksheet_title])

    # If the columns are different, overwrite the entire data
    if existing_columns != column_names:
      self._overwrite_worksheet(worksheet_title, data_to_write, n_index_cols)
      return

    # Otherwise only update rows that have changed
    update_batches = self._construct_update_batches(
        data_to_write, existing_data
    )
    self._update_size_of_worksheet(worksheet_title, data_to_write)
    self.spreadsheet.worksheet(worksheet_title).batch_update(update_batches)
    self._update_worksheet_formatting(worksheet_title, n_index_cols)

  def delete_worksheet(self, worksheet_title: str) -> None:
    """Deletes the worksheet with the given title."""
    worksheet = self.spreadsheet.worksheet(worksheet_title)
    self.spreadsheet.del_worksheet(worksheet)


class GoogleSheetsLogSender:
  """A class for sending logs to Google Sheets.

  This class provides a convenient way to send logs to Google Sheets. It creates
  a new worksheet or uses an existing one to store the logs.

  Attributes:
    client: The Google Sheets client.
    spreadsheet: The Google Sheets spreadsheet.
    log_worksheet: The Google Sheets worksheet for logs.
  """

  HEADINGS = ["UTC Timestamp", "Log Level", "Logger Name", "Message"]

  def __init__(
      self,
      sheet_url: str,
      log_worksheet_name: str = "Logs",
  ):
    """Initializes the GoogleSheetsLogSender object.

    Args:
      sheet_url: The URL of the spreadsheet to send logs to.
      log_worksheet_name: The name of the worksheet to send logs to. If the
        worksheet does not exist, then it is created.
    """
    self.client = get_gspread_client()
    self.spreadsheet = self.client.open_by_url(sheet_url)

    existing_worksheets = [
        sheet.title for sheet in self.spreadsheet.worksheets()
    ]
    if log_worksheet_name not in existing_worksheets:
      self.log_worksheet = self.spreadsheet.add_worksheet(
          title=log_worksheet_name, rows=2, cols=4
      )
      self.log_worksheet.update([self.HEADINGS], raw=False)
      self.log_worksheet.format(ranges=["A1:D1"], format=HEADING_FORMAT)
    else:
      self.log_worksheet = self.spreadsheet.worksheet(log_worksheet_name)

    actual_headings = self.log_worksheet.row_values(1)
    if actual_headings != self.HEADINGS:
      raise ValueError(
          "The first row of the log worksheet must be the expected headings:"
          f" {self.HEADINGS}, but got {actual_headings}."
      )

  def write_log(self, msg: logging.LogRecord) -> None:
    """Writes a log record to the log worksheet.

    Args:
      msg: The log record to write.
    """
    self.log_worksheet.insert_row(
        [
            time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(msg.created)),
            msg.levelname,
            msg.name,
            msg.getMessage(),
        ],
        index=2,
    )


class GoogleSheetsHandler(logging.Handler):
  """A logging handler that sends logs to Google Sheets.

  This class is a logging handler that sends logs to Google Sheets. It creates
  a new worksheet or uses an existing one to store the logs.

  Attributes:
    sender: The GoogleSheetsLogSender object.
  """

  def __init__(
      self,
      sheet_url: str,
      log_worksheet_name: str = "Logs",
  ) -> None:
    """Initializes the GoogleSheetsHandler object.

    Args:
      sheet_url: The URL of the spreadsheet to send logs to.
      log_worksheet_name: The name of the worksheet to send logs to. If the
        worksheet does not exist, then it is created.
    """
    self.sender = GoogleSheetsLogSender(
        sheet_url=sheet_url,
        log_worksheet_name=log_worksheet_name,
    )
    super().__init__()

  def emit(self, record: logging.LogRecord) -> None:
    """Emits a log record to the log worksheet.

    Args:
      record: The log record to emit.
    """
    self.sender.write_log(record)


def create_template_copycat_sheet(
    spreadsheet_name: str = "Copycat Template", include_demo_data: bool = True
) -> str:
  """Creates a template Copycat sheet.

  The template sheet contains worksheets for existing ads, new keywords, and
  extra instructions. This sheet is the correct format for the Copycat UI.

  If include_demo_data is True, then the template sheet will contain demo data
  for existing ads, new keywords, and extra instructions. If include_demo_data
  is False, then the template sheet will only contain column headings.

  The template sheet contains the following worksheets and column headings:
  - Existing Ads:
    - Campaign ID
    - Ad Group
    - URL
    - Ad Strength
    - Keywords
    - Headline 1
    - ...
    - Headline 15
    - Description 1
    - ...
    - Description 4
  - New Keywords:
    - Campaign ID
    - Ad Group
    - Keyword
  - Extra Instructions for New Ads:
    - Campaign ID
    - Ad Group
    - Version
    - Extra Instructions


  Args:
    spreadsheet_name: The name of the spreadsheet to create.
    include_demo_data: Whether to include demo data in the template sheet.

  Returns:
    The URL of the template sheet.
  """
  sheet = GoogleSheet.new(spreadsheet_name)

  demo_data_path = pathlib.Path(__file__).parent / "demo_data"

  demo_data_files = demo_data_path.glob("*.csv")
  data = {
      demo_data_file.stem: pd.read_csv(demo_data_file)
      for demo_data_file in demo_data_files
  }

  if not include_demo_data:
    for name in data.keys():
      data[name] = pd.DataFrame(columns=data[name].columns)

  sheet[TEMPLATE_EXISTING_ADS_WORKSHEET_NAME] = (
      data["existing_ads"].set_index(["Campaign ID", "Ad Group"]).fillna("")
  )
  sheet[TEMPLATE_NEW_KEYWORDS_WORKSHEET_NAME] = (
      data["new_keywords"].set_index(["Campaign ID", "Ad Group"]).fillna("")
  )
  sheet[TEMPLATE_EXTRA_INSTRUCTIONS_WORKSHEET_NAME] = (
      data["extra_instructions"]
      .set_index(["Campaign ID", "Ad Group", "Version"])
      .fillna("")
  )

  sheet.delete_worksheet("Sheet1")
  return sheet.url



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for working with data."""

import itertools
import logging
from typing import Any, Callable, Generator

import numpy as np
import pandas as pd


LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def collapse_headlines_and_descriptions(
    data: pd.DataFrame,
) -> pd.DataFrame:
  """Collapses headline and description columns into two new columns.

  Assumes that the headline and description column names have the format
  "Headline {i}" and "Description {i}", where {i} is the headline number, with
  a single headline or description per column.

  Then collapses them into two new columns, "headlines" and "descriptions",
  containing lists of headlines and descriptions respectively.

  Args:
    data: The input DataFrame.

  Returns:
    A new DataFrame with the headline and description columns collapsed into two
    new columns, "headlines" and "descriptions".
  """
  output_data = data.copy()

  headline_cols = [
      c
      for c in output_data.columns
      if c.startswith("Headline ")
      and c.split(" ")[1].isdigit()
      and len(c.split(" ")) == 2
  ]
  description_cols = [
      c
      for c in output_data.columns
      if c.startswith("Description ")
      and c.split(" ")[1].isdigit()
      and len(c.split(" ")) == 2
  ]
  output_data["headlines"] = pd.Series(
      {
          k: list(
              filter(lambda x: x != "--" and x, v),
          )
          for k, v in output_data[headline_cols].T.to_dict("list").items()
      },
      index=output_data.index,
  )
  output_data["descriptions"] = pd.Series(
      {
          k: list(filter(lambda x: x != "--" and x, v))
          for k, v in output_data[description_cols].T.to_dict("list").items()
      },
      index=output_data.index,
  )

  output_data = output_data.drop(columns=headline_cols + description_cols)

  return output_data


def _explode_to_columns(
    output_name: str, max_columns: int | None = None
) -> Callable[[list[Any]], pd.Series]:
  """Returns a function that explodes a list into a Series of columns.

  The returned function takes a list as input and returns a Series where each
  element of the list is assigned to a separate column, with the column names
  having the format f"{output_name} {i}", where i is the index of the element
  in the list and starts at 1.

  Args:
    output_name: The name of the output columns.
    max_columns: The maximum number of columns to create. If None, then the
      number of columns will be equal to the length of the input list.

  Returns:
    A function that explodes a list into a Series of columns.

  Raises:
    ValueError: If the input to the function is not a list or if the length of
    the list is greater than the maximum number of columns.
  """

  def apply_explode_to_columns(list_col: list[Any]) -> pd.Series:
    nonlocal max_columns
    nonlocal output_name
    if max_columns is None:
      max_columns = len(list_col)
    elif len(list_col) > max_columns:
      raise ValueError(
          "The input to the explode_to_columns function must be a list of"
          f" length {max_columns} or less, got {len(list_col)} instead."
      )

    if not isinstance(list_col, list):
      raise ValueError(
          "The input to the explode_to_columns function must be a list, got"
          f" {type(list_col)} instead."
      )
    return pd.Series(
        list_col + ["--" for _ in range(max_columns - len(list_col))],
        index=[f"{output_name} {i+1}" for i in range(max_columns)],
    )

  return apply_explode_to_columns


def explode_headlines_and_descriptions(
    data: pd.DataFrame,
    max_headlines: int | None = None,
    max_descriptions: int | None = None,
) -> pd.DataFrame:
  """Explodes headline and description columns into separate columns.

  Assumes that the headline and description columns have been collapsed into two
  new columns, "headlines" and "descriptions", containing lists of headlines and
  descriptions respectively.

  Then explodes them into separate columns, with the column names having the
  format "Headline {i}" and "Description {i}", where {i} is the index of the
  headline or description and starts at 1.

  Args:
    data: The input DataFrame.
    max_headlines: The maximum number of headlines to explode. If None, then the
      number of headlines will be equal to the length of the input list.
    max_descriptions: The maximum number of descriptions to explode. If None,
      then the number of descriptions will be equal to the length of the input
      list.

  Returns:
    A new DataFrame with the headline and description columns exploded into
    separate columns.

  Raises:
    ValueError: If the index of the input data is not unique.
  """
  if not data.index.is_unique:
    raise ValueError(
        "The index of the input data is not unique, cannot explode headlines"
        " and descriptions."
    )

  if "headlines" in data:
    headlines = (
        data["headlines"]
        .apply(_explode_to_columns("Headline", max_headlines))
        .fillna("--")
    )
  elif max_headlines is not None:
    headlines = data.apply(
        lambda x: _explode_to_columns("Headline", max_headlines)([]), axis=1
    ).fillna("--")
  else:
    headlines = pd.DataFrame()

  if "descriptions" in data:
    descriptions = (
        data["descriptions"]
        .apply(_explode_to_columns("Description", max_descriptions))
        .fillna("--")
    )
  elif max_descriptions is not None:
    descriptions = data.apply(
        lambda x: _explode_to_columns("Description", max_descriptions)([]),
        axis=1,
    ).fillna("--")
  else:
    descriptions = pd.DataFrame()

  output_data = (
      data.copy()
      .drop(columns=["headlines", "descriptions"])
      .merge(headlines, left_index=True, right_index=True)
      .merge(descriptions, left_index=True, right_index=True)
  )

  return output_data


def iterate_over_batches(
    data: pd.DataFrame, batch_size: int, limit_rows: int | None = None
) -> Generator[pd.DataFrame, None, None]:
  """Iterates over batches of data.

  Args:
    data: The input DataFrame.
    batch_size: The size of each batch.
    limit_rows: The maximum number of rows to iterate over. If None, all rows
      will be iterated over.

  Yields:
    A generator that yields batches of data.
  """
  if limit_rows is None or limit_rows > len(data):
    limit_rows = len(data)

  n_regular_batches = limit_rows // batch_size
  final_batch_size = limit_rows % batch_size

  for i in range(n_regular_batches):
    yield data.iloc[i * batch_size : (i + 1) * batch_size]

  if final_batch_size:
    yield data.iloc[
        n_regular_batches * batch_size : n_regular_batches * batch_size
        + final_batch_size
    ]


def _join_additional_instructions_data(
    additional_instructions_data: pd.DataFrame,
    target_data: pd.DataFrame,
    additional_instructions_column: str,
) -> pd.DataFrame:
  """Prepares the additional instructions data for merging.

  The additional instructions are joined onto the target data on the join
  columns. If the value of a join column in the additional instructions data is
  set to __ALL__, then it will be merged with all values of that column in the
  target data. If there are duplicates in the additional instructions data then
  they are joined together on separate lines.

  Args:
    additional_instructions_data: The additional instructions data.
    target_data: The target data to join the additional instructions data to.
    additional_instructions_column: The column in the additional instructions
      data that contains the additional instructions.

  Returns:
    A new DataFrame with the additional instructions data joined onto the target
    data.
  """
  join_columns = list(additional_instructions_data.index.names)
  complete_index = target_data.reset_index()[join_columns].copy()
  additional_instructions_data = additional_instructions_data.reset_index()

  exploded_additional_instructions = []
  for n_join_columns in range(len(join_columns) + 1):
    for selected_join_columns in itertools.combinations(
        join_columns, n_join_columns
    ):
      selected_join_columns = list(selected_join_columns)
      unselected_join_columns = [
          col for col in join_columns if col not in selected_join_columns
      ]
      mask = np.array([True] * len(additional_instructions_data))
      if selected_join_columns:
        mask &= np.all(
            additional_instructions_data[selected_join_columns].values
            != "__ALL__",
            axis=1,
        )
      if unselected_join_columns:
        mask &= np.all(
            additional_instructions_data[unselected_join_columns].values
            == "__ALL__",
            axis=1,
        )

      if not mask.any():
        continue

      data_to_merge = additional_instructions_data.loc[
          mask, selected_join_columns + [additional_instructions_column]
      ]
      if selected_join_columns:
        exploded_additional_instructions.append(
            complete_index.merge(
                data_to_merge,
                on=selected_join_columns,
                how="inner",
            )
        )
      else:
        exploded_additional_instructions.append(
            complete_index.merge(
                data_to_merge,
                how="cross",
            )
        )

  final_additional_instructions = (
      pd.concat(exploded_additional_instructions)
      .groupby(join_columns)
      .agg(
          additional_instructions=(
              additional_instructions_column,
              lambda x: "\n".join(sorted(list(set(x)))),
          )
      )
  )
  merged_data = target_data.join(
      final_additional_instructions,
      how="left",
  )
  merged_data["additional_instructions"] = merged_data[
      "additional_instructions"
  ].fillna("")
  return merged_data


def construct_generation_data(
    *,
    new_keywords_data: pd.DataFrame,
    additional_instructions_data: pd.DataFrame | None = None,
    existing_generations_data: pd.DataFrame | None = None,
    keyword_column: str = "keyword",
    additional_instructions_column: str = "additional_instructions",
    existing_headlines_column: str = "existing_headlines",
    existing_descriptions_column: str = "existing_descriptions",
    version_column: str = "version",
    n_versions: int = 1,
) -> pd.DataFrame:
  """Constructs the generation data.

  The new keywords data should contain the following columns:
    - The columns in the index_columns list.
    - The column specified by the keyword_column, containing a single keyword
      per row.

  (Optionl) The additional instructions data should contain the following
  columns:
    - The columns in the index_columns list.
    - The column specified by the version_column, containing the version number
    of the ad.
    - The column specified by the additional_instructions_column, containing a
      single additional instruction per row.
  For the additional instructions data, if any of the values of the index
  columns or the version column is set to __ALL__, then it will be merged with
  all values of those columns in the target data.

  (Optional) The existing generations data should contain the following columns:
    - The columns in the index_columns list.
    - The column specified by the version_column, containing the version number
    of the ad.
    - The column specified by the existing_headlines_column, containing a list
      of previously generated headlines.
    - The column specified by the existing_descriptions_column, containing a
      list of previously generated descriptions.
  For the existing generations data, the index and versions columns must be
  unique.

  Args:
    new_keywords_data: The new keywords data.
    additional_instructions_data: The additional instructions data.
    existing_generations_data: The existing generations data.
    keyword_column: The column in the new keywords data that contains the
      keywords.
    additional_instructions_column: The column in the additional instructions
      data that contains the additional instructions.
    existing_headlines_column: The column in the existing generations data that
      contains the existing headlines.
    existing_descriptions_column: The column in the existing generations data
      that contains the existing descriptions.
    version_column: The column in the additional instructions data and the
      existing generations data that contains the version number of the ad.
    n_versions: The number of versions to generate for each new ad.

  Returns:
    A new DataFrame with the generation data.

  Raises:
    ValueError: If the index columns are not unique in the existing generations
    data.
  """
  index_columns = list(new_keywords_data.index.names)
  join_columns = index_columns + [version_column]

  generation_data = (
      new_keywords_data.reset_index()
      .groupby(index_columns)
      .agg(keywords=(keyword_column, ", ".join))
      .reset_index()
  )
  generation_data[version_column] = [list(range(1, n_versions + 1))] * len(
      generation_data
  )

  generation_data = generation_data.explode(version_column, ignore_index=True)
  generation_data[version_column] = generation_data[version_column].astype(str)
  generation_data.set_index(join_columns, inplace=True)

  if (
      existing_generations_data is not None
      and len(existing_generations_data) > 0
  ):
    if set(existing_generations_data.index.names) != set(join_columns):
      error_message = (
          "The index columns of the existing_generations_data do not match the"
          f" expected columns, expected: {join_columns}, got:"
          f" {existing_generations_data.index.names}"
      )
      LOGGER.error(error_message)
      raise ValueError(error_message)

    if not existing_generations_data.index.is_unique:
      error_message = (
          "The index columns of the existing_generations_data are not unique,"
          " cannot merge with the new keywords data."
      )
      LOGGER.error(error_message)
      raise ValueError(error_message)

    generation_data = generation_data.join(
        existing_generations_data[
            [existing_headlines_column, existing_descriptions_column]
        ],
        how="left",
    ).rename(
        columns={
            existing_headlines_column: "existing_headlines",
            existing_descriptions_column: "existing_descriptions",
        }
    )
    generation_data[["existing_headlines", "existing_descriptions"]] = (
        generation_data[["existing_headlines", "existing_descriptions"]]
        .fillna("")
        .map(list)
    )
  else:
    generation_data["existing_headlines"] = ""
    generation_data["existing_descriptions"] = ""
    generation_data[["existing_headlines", "existing_descriptions"]] = (
        generation_data[["existing_headlines", "existing_descriptions"]].map(
            list
        )
    )

  if (
      additional_instructions_data is not None
      and len(additional_instructions_data) > 0
  ):
    if set(additional_instructions_data.index.names) != set(join_columns):
      error_message = (
          "The index columns of the additional_instructions_data do not match"
          f" the expected columns, expected: {join_columns}, got:"
          f" {additional_instructions_data.index.names}"
      )
      LOGGER.error(error_message)
      raise ValueError(error_message)

    generation_data = _join_additional_instructions_data(
        additional_instructions_data=additional_instructions_data[
            [additional_instructions_column]
        ],
        target_data=generation_data,
        additional_instructions_column=additional_instructions_column,
    )
  else:
    generation_data["additional_instructions"] = ""

  return generation_data


def explode_generated_ad_object(
    data: pd.DataFrame,
    *,
    generated_ad_object_column="generated_ad_object",
    headlines_column="headlines",
    descriptions_column="descriptions",
    success_column="Success",
    headline_count_column="Headline Count",
    description_count_column="Description Count",
    headlines_are_memorised_column="Headlines are Memorized",
    descriptions_are_memorised_column="Descriptions are Memorized",
    style_similarity_column="Style Similarity",
    keyword_similarity_column="Keyword Similarity",
    warning_message_column="Warnings",
    error_message_column="Errors",
):
  """Explodes the generated_ad_object column into separate columns.

  Args:
    data: The input DataFrame.
    generated_ad_object_column: The column in the data that contains the
      generated ad objects.
    headlines_column: The name of the column to store the headlines.
    descriptions_column: The name of the column to store the descriptions.
    success_column: The name of the column to store whether the generation was
      successful.
    headline_count_column: The name of the column to store the number of
      headlines.
    description_count_column: The name of the column to store the number of
      descriptions.
    headlines_are_memorised_column: The name of the column to store whether the
      headlines are memorized.
    descriptions_are_memorised_column: The name of the column to store whether
      the descriptions are memorized.
    style_similarity_column: The name of the column to store the style
      similarity.
    keyword_similarity_column: The name of the column to store the keyword
      similarity.
    warning_message_column: The name of the column to store the warning
      messages.
    error_message_column: The name of the column to store the error messages.

  Returns:
    A new DataFrame with the generated_ad_object column exploded into separate
    columns.
  """
  data = data.copy()
  data[headlines_column] = data[generated_ad_object_column].apply(
      lambda x: x.google_ad.headlines
  )
  data[descriptions_column] = data[generated_ad_object_column].apply(
      lambda x: x.google_ad.descriptions
  )
  data[success_column] = data[generated_ad_object_column].apply(
      lambda x: x.success
  )
  data[headlines_are_memorised_column] = data[generated_ad_object_column].apply(
      lambda x: x.evaluation_results.headlines_are_memorised
  )
  data[descriptions_are_memorised_column] = data[
      generated_ad_object_column
  ].apply(lambda x: x.evaluation_results.descriptions_are_memorised)
  data[style_similarity_column] = data[generated_ad_object_column].apply(
      lambda x: x.evaluation_results.style_similarity
  )
  data[keyword_similarity_column] = data[generated_ad_object_column].apply(
      lambda x: x.evaluation_results.keyword_similarity
  )
  data[warning_message_column] = data[generated_ad_object_column].apply(
      lambda x: x.warning_message
  )
  data[error_message_column] = data[generated_ad_object_column].apply(
      lambda x: x.error_message
  )
  data[headline_count_column] = data[generated_ad_object_column].apply(
      lambda x: x.google_ad.headline_count
  )
  data[description_count_column] = data[generated_ad_object_column].apply(
      lambda x: x.google_ad.description_count
  )

  return data



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains the sub-page for loading data into Copycat and validating it."""

import mesop as me

from copycat.ui import components
from copycat.ui import event_handlers
from copycat.ui import states


def prepare_data() -> None:
  """Creates the sub-page for loading data into Copycat and validating it."""
  state = me.state(states.AppState)
  with components.column(width="100%"):
    with components.rounded_box_section(
        title="Validate Google Sheet", width="100%"
    ):
      me.text(
          "The data in the Google Sheet must contain three worksheets,"
          " 'Training Ads', 'New Keywords' and 'Extra Instructions for New"
          " Ads', and they must follow the expected format exactly. This is"
          " especially important to check if you have created the Google Sheet"
          " manually. Make sure all the column names are correct and in the"
          " correct order, and that the correct number of columns and rows are"
          " frozen in the google sheet. To see what is expected, select New /"
          " Load and then create a new Google Sheet and check the columns it"
          " contains."
      )
      me.text(
          "If the validation fails, you should check the Logs tab of your"
          " Google Sheet to see which rows are invalid and why.",
          style=me.Style(margin=me.Margin(top=15)),
      )

      with components.row(align_items="center"):
        me.button(
            label="Validate Sheet",
            type="flat",
            on_click=event_handlers.validate_sheet,
            style=me.Style(margin=me.Margin.symmetric(vertical=20)),
        )
        if state.google_sheet_is_valid:
          me.icon("check_circle")
          me.text("Validated")
        else:
          me.icon("warning", style=me.Style(color=me.theme_var("error")))
          me.text("Not Validated", style=me.Style(color=me.theme_var("error")))



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
import mock
import numpy as np
from sklearn import cluster

from copycat import keyword_organiser


class KeywordOrganiserTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.train_keyword_embeddings = np.array([
        [0.5, 0.7, 1.0],
        [0.2, 0.9, 3.0],
        [-1.0, -2.0, -1.0],
        [-1.0, -5.0, -2.0],
    ])
    self.train_target_ad_groups = np.array([1, 1, 0, 0])
    self.test_keyword_embeddings = np.array([
        [-0.5, 0.2, 2.0],
        [0.3, 0.1, -3.5],
        [-1.0, 2.0, -1.0],
    ])

  def test_fit_learns_the_distance_threshold(self):
    clusterer = keyword_organiser.BirchAgglomerativeKeywordClusterer()
    clusterer.fit(
        keyword_embeddings=self.train_keyword_embeddings,
        target_ad_groups=self.train_target_ad_groups,
    )
    self.assertAlmostEqual(clusterer.distance_threshold, 3.1622776601683795)

  def test_predict_raises_error_if_not_fit(self):
    clusterer = keyword_organiser.BirchAgglomerativeKeywordClusterer()
    with self.assertRaisesWithLiteralMatch(
        ValueError, "Must call fit before predicting ad groups."
    ):
      clusterer.predict(self.test_keyword_embeddings)

  def test_predict_returns_expected_clusters(self):
    clusterer = keyword_organiser.BirchAgglomerativeKeywordClusterer()
    clusterer.fit(
        keyword_embeddings=self.train_keyword_embeddings,
        target_ad_groups=self.train_target_ad_groups,
    )
    predicted_ad_groups = clusterer.predict(self.test_keyword_embeddings)
    np.testing.assert_array_equal(predicted_ad_groups, np.array([1, 2, 0]))

  def test_evaluate_returns_expected_score(self):
    clusterer = keyword_organiser.BirchAgglomerativeKeywordClusterer()
    clusterer.fit(
        keyword_embeddings=self.train_keyword_embeddings,
        target_ad_groups=self.train_target_ad_groups,
    )
    score = clusterer.evaluate(
        keyword_embeddings=self.train_keyword_embeddings,
        target_ad_groups=self.train_target_ad_groups,
    )
    self.assertAlmostEqual(score, 1.0)

  def test_fit_calls_agglomerative_clustering_with_expected_args(
      self,
  ):
    clusterer = keyword_organiser.BirchAgglomerativeKeywordClusterer()
    with mock.patch(
        "sklearn.cluster.AgglomerativeClustering"
    ) as mock_agglomerative_clustering:
      mock_agglomerative_clustering.return_value.distances_ = np.array([
          1.0,
          2.0,
          3.0,
          4.0,
      ])
      mock_agglomerative_clustering.return_value.children_ = [
          (0, 1),
          (2, 3),
      ]

      clusterer.fit(
          keyword_embeddings=self.train_keyword_embeddings,
          target_ad_groups=self.train_target_ad_groups,
      )
      mock_agglomerative_clustering.assert_called_once_with(
          n_clusters=None,
          linkage="average",
          distance_threshold=10.0,
          compute_distances=True,
          compute_full_tree=True,
      )

  def test_predict_calls_agglomerative_clustering_with_expected_args(
      self,
  ):
    clusterer = keyword_organiser.BirchAgglomerativeKeywordClusterer()
    clusterer.fit(
        keyword_embeddings=self.train_keyword_embeddings,
        target_ad_groups=self.train_target_ad_groups,
    )
    with mock.patch(
        "sklearn.cluster.AgglomerativeClustering",
    ) as mock_agglomerative_clustering:
      clusterer.predict(self.test_keyword_embeddings)
      mock_agglomerative_clustering.assert_called_once_with(
          n_clusters=None,
          linkage="average",
          distance_threshold=clusterer.distance_threshold,
      )

  def test_birch_used_if_scale_is_greater_than_zero(self):
    clusterer = keyword_organiser.BirchAgglomerativeKeywordClusterer(
        birch_scale=0.5
    )
    clusterer.fit(
        keyword_embeddings=self.train_keyword_embeddings,
        target_ad_groups=self.train_target_ad_groups,
    )
    with mock.patch(
        "sklearn.cluster.Birch",
    ) as mock_birch:
      with mock.patch(
          "sklearn.cluster.AgglomerativeClustering",
      ) as mock_agglomerative_clustering:
        predicted_ad_groups = clusterer.predict(self.test_keyword_embeddings)

        # Test birch is instantiated with the expected arguments
        mock_agglomerative_clustering_instance = (
            mock_agglomerative_clustering.return_value
        )
        mock_birch.assert_called_once_with(
            threshold=clusterer.distance_threshold * clusterer.birch_scale,
            n_clusters=mock_agglomerative_clustering_instance,
        )

        # Test birch fit_predict is called with the expected arguments
        mock_birch_instance = mock_birch.return_value
        mock_birch_instance.fit_predict.assert_called_once_with(
            self.test_keyword_embeddings
        )

        # Test fit predict returns the output from fit predict
        self.assertIs(
            predicted_ad_groups, mock_birch_instance.fit_predict.return_value
        )

  def test_birch_not_used_if_scale_is_zero(self):
    clusterer = keyword_organiser.BirchAgglomerativeKeywordClusterer(
        birch_scale=0.0
    )
    clusterer.fit(
        keyword_embeddings=self.train_keyword_embeddings,
        target_ad_groups=self.train_target_ad_groups,
    )
    with mock.patch(
        "sklearn.cluster.Birch",
    ) as mock_birch:
      with mock.patch(
          "sklearn.cluster.AgglomerativeClustering",
      ) as mock_agglomerative_clustering:
        predicted_ad_groups = clusterer.predict(self.test_keyword_embeddings)

        # Test birch is not called
        mock_birch.assert_not_called()

        # Test agglomerative fit_predict is called with the expected arguments
        mock_agglomerative_clustering_instance = (
            mock_agglomerative_clustering.return_value
        )
        mock_agglomerative_clustering_instance.fit_predict.assert_called_once_with(
            self.test_keyword_embeddings
        )

        # Test fit predict returns the output from fit predict
        self.assertIs(
            predicted_ad_groups,
            mock_agglomerative_clustering_instance.fit_predict.return_value,
        )

  @parameterized.parameters([
      1.0,
      -0.1,
  ])
  def test_value_error_raised_if_birch_scale_is_invalid(self, birch_scale):
    with self.assertRaisesWithLiteralMatch(
        ValueError, f"Birch scale must be >=0 and <1, but got {birch_scale}"
    ):
      keyword_organiser.BirchAgglomerativeKeywordClusterer(
          birch_scale=birch_scale
      )


if __name__ == "__main__":
  absltest.main()
